# SPDX-License-Identifier: Apache-2.0
"""Package for exporters."""
from nv_one_logger.exporter.exporter import Exporter  # noqa: F401
from nv_one_logger.exporter.file_exporter import FileExporter  # noqa: F401
from nv_one_logger.exporter.logger_exporter import LoggerExporter  # noqa: F401



# SPDX-License-Identifier: Apache-2.0
import traceback
import uuid
from collections import OrderedDict
from contextvars import ContextVar
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Callable, Dict, List, Optional

from overrides import override  # type: ignore[ancereportUnknownVariableType]

from nv_one_logger.api.recorder import Recorder
from nv_one_logger.core.attributes import Attributes
from nv_one_logger.core.event import ErrorEvent, Event, TelemetryDataError
from nv_one_logger.core.exceptions import OneLoggerError, assert_that
from nv_one_logger.core.internal.logging import get_logger
from nv_one_logger.core.internal.safe_execution import safely_execute
from nv_one_logger.core.span import Span, SpanName
from nv_one_logger.core.time import TracingTimestamp
from nv_one_logger.exporter.exporter import Exporter

_logger = get_logger(__name__)

# This constant defines the number of failures after which an exporter is disabled. This allows us to avoid unnecessary overhead of calling an exporter
# when it is fatally broken and the data on the backend is not useful due to being incomplete.
_DISABLE_EXPORTER_AFTER_N_FAILURES: int = 5


class ExportCustomizationMode(Enum):
    """The mode of exporting spans (and their associated events and attribytes) to exporters."""

    EXPORT_ALL_SPANS = auto()  # Export all spans and their associated events and attributes to all exporters.
    WHITELIST_SPANS = auto()  # Export only spans that match the whitelist and their associated events and attributes to all exporters.
    BLACKLIST_SPANS = auto()  # Export all spans except those that match the blacklist and their associated events and attributes to all exporters.


@dataclass
class _ExporterState:
    # Number of failed exports so far.
    num_errors: int = 0

    # True means at least some data about the current execution of the application has been successfully sent to this exporter.
    any_data_exported_successfully: bool = False

    # Determines if we force disabled an exporter due to seeing too many consecutive failures.
    force_disabled: bool = False

    # Determines if we have already reported a telemetry data error for this exporter.
    # This variable is used to ensure that we only report telemetry data errors once for each exporter.
    # Note that this could become true either when the exporter encounters an error or when the
    # recorder.telemetry_data_error() method is called by another component of the library.
    telemetry_data_error_reported: bool = False


class DefaultRecorder(Recorder):
    """
    The default recorder implementation of the :class:`one_logger.core.recorder.Recorder` interface.

    This implementation
    - supports exporting to multiple backends.
    - Tracks the parent-child relationship between spans: if a span is started while another span is active, it will be created
      as a child of the active span (unless a parent span is specified explicitly).
    - does NOT filter any spans, events, or attributes nor adds any attributes to the spans or events automatically. However,
      it supports am optional whitelist or blacklist filter for spams to export.
    - It handles errors from exporters differently from other errors. The reason behind this design decision is that
      exporter errors do not necessarily mean that the library has a fatal bug. An intermittent server error could
      cause one exporter to fail while all other exportes may still be working. So for exporter errors, this recorder
        - Catches the error, adds an error log, and records a telemetry data error on the corresponding backend only
          but does not propagate the error to the caller.
        - It disables an exporter after a certain number of failures. This is a heuristic to avoid unnecessary overhead of calling an exporter
          when it is fatally broken and the data on the backend is not useful due to being incomplete.


    It is possible to create a DefaultRecorder with no backends, in which case it will not record anything,
    this is useful if the application is running on thousands of GPUs, in which case the overhead of recording
    from each GPU is too high, and we only want to record from rank zero or a subset of ranks.
    """

    def __init__(
        self,
        exporters: List[Exporter],
        export_customization_mode: ExportCustomizationMode = ExportCustomizationMode.EXPORT_ALL_SPANS,
        span_name_filter: Optional[List[SpanName]] = None,
    ):
        """Initialize the DefaultRecorder with a list of exporters.

        Args:
            exporters: A list of exporters to use.
            export_customization_mode: The mode of exporting spans (and their associated events and attribytes) to exporters.
            span_name_filter: This argument should be interpretted wrt the value of export_customization_mode:
                If export_customization_mode is ExportCustomizationMode.EXPORT_ALL_SPANS, span_name_filter should not be set.
                If export_customization_mode is ExportCustomizationMode.WHITELIST_SPANS, span_name_filter is a list of span names to export (whitelist).
                If export_customization_mode is ExportCustomizationMode.BLACKLIST_SPANS, span_name_filter is a list of span names to not export (blacklist).
        """
        self._exporters = exporters
        # Keep track of the state of each exporter so that we can handle export errors gracefully.
        self._exporter_state: Dict[Exporter, _ExporterState] = {exporter: _ExporterState() for exporter in exporters}

        self._export_customization_mode = export_customization_mode
        self._span_name_filter = span_name_filter

        if export_customization_mode == ExportCustomizationMode.EXPORT_ALL_SPANS:
            assert_that(
                span_name_filter is None, "span_name_filter should not be set when export_customization_mode is ExportCustomizationMode.EXPORT_ALL_SPANS"
            )
        elif export_customization_mode in {ExportCustomizationMode.WHITELIST_SPANS, ExportCustomizationMode.BLACKLIST_SPANS}:
            assert_that(span_name_filter, f"span_name_filter should be a non-empty list when export_customization_mode is {export_customization_mode.name}")
        else:
            raise OneLoggerError(f"Invalid export_customization_mode: {export_customization_mode}")

        # Keeps track of the latest active span in the current context (thread or task). That is, the span that was most recently
        # started and not yet stopped.
        # This is used as the parent span for a newly created span (when no parent span is provided explicitly).
        self._latest_active_span: ContextVar[Optional[Span]] = ContextVar("latest_active_span", default=None)

        self._spans: OrderedDict[uuid.UUID, Span] = OrderedDict()
        if len(self._exporters) > 0:
            _logger.debug(f"Initializing DefaultRecorder with {len(self._exporters)} exporters.")
        else:
            _logger.debug("Initializing DefaultRecorder with no exporters, exporting is disabled")
        for exporter in self._exporters:
            exporter.initialize()

        self._closed = False

    @safely_execute
    @override
    def start(
        self,
        span_name: SpanName,
        span_attributes: Optional[Attributes] = None,
        start_event_attributes: Optional[Attributes] = None,
        start_time: Optional[TracingTimestamp] = None,
        parent_span: Optional[Span] = None,
    ) -> Span:
        """
        Create a new span and record its start event.

        Default implementation of Recorder.start() exporting to all exporters of the recorder.
        See the docstrings of the Recorder interface for more details.
        """
        assert_that(not self._closed, "Recorder is already closed.")
        span = Span.create(
            name=span_name,
            parent_span=parent_span if parent_span else self._latest_active_span.get(),
            span_attributes=span_attributes,
            start_event_attributes=start_event_attributes,
            start_time=start_time,
        )
        self._spans[span.id] = span
        self._latest_active_span.set(span)
        if self._should_export_span(span):
            for exporter in self._exporters:
                self._execute_export(exporter, exporter.export_start, span)
        return span

    @safely_execute
    @override
    def stop(
        self,
        span: Span,
        stop_event_attributes: Optional[Attributes] = None,
        stop_time: Optional[TracingTimestamp] = None,
    ) -> None:
        """Stop the span and record its stop event.

        Default implementation of Recorder.stop() exporting to all exporters of the recorder.
        See the docstrings of the Recorder interface for more details.
        """
        assert_that(not self._closed, "Recorder is already closed.")
        assert_that(span, "No span to stop")
        try:
            self._spans.pop(span.id)
        except KeyError:
            raise OneLoggerError(f"Span {span.id} not found")

        span.stop(stop_event_attributes, stop_time)

        # If the span that is being stopped is the latest active span, we need to update the latest active span to the parent of the span.
        if self._latest_active_span.get() and span.id == self._latest_active_span.get().id:
            # Normally, that parent of the current active span should be active, but let's err on the side of caution and check.
            if span.parent_span and not span.parent_span.active:
                _logger.error(
                    f"The parent of span {span.name_str} (id={span.id}) is stopped before its child span. "
                    f"The parent span is {span.parent_span.name_str} (id={span.parent_span.id})"
                )
            new_latest_active_span = span.parent_span
            while new_latest_active_span and not new_latest_active_span.active:
                new_latest_active_span = new_latest_active_span.parent_span
            self._latest_active_span.set(new_latest_active_span)

        if self._should_export_span(span):
            for exporter in self._exporters:
                self._execute_export(exporter, exporter.export_stop, span)

    @safely_execute
    @override
    def event(self, span: Span, event: Event) -> Event:
        """Add an event to a span and record it.

        Default implementation of Recorder.event() exporting to all exporters of the recorder.
        See the docstrings of the Recorder interface for more details.
        """
        assert_that(not self._closed, "Recorder is already closed.")
        assert_that(span, "No span to add event to")
        span.add_event(event)

        if self._should_export_span(span):
            for exporter in self._exporters:
                self._execute_export(exporter, exporter.export_event, event, span)
        return event

    @safely_execute
    @override
    def error(
        self,
        span: Span,
        error_message: str,
        exception: Optional[Exception] = None,
    ) -> ErrorEvent:
        """Add an error to a span and record it.

        Default implementation of Recorder.error() exporting to all exporters of the recorder.
        See the docstrings of the Recorder interface for more details.
        """
        assert_that(not self._closed, "Recorder is already closed.")
        assert_that(span, "No span to add error to")

        event = ErrorEvent.create(
            error_message=error_message,
            exception=exception,
            exception_traceback=traceback.format_exc() if exception else None,
        )
        span.add_event(event)
        if self._should_export_span(span):
            for exporter in self._exporters:
                self._execute_export(exporter, exporter.export_error, event, span)
        return event

    @safely_execute
    @override
    def telemetry_data_error(self, error_message: str, attributes: Optional[Attributes] = None) -> None:
        """Report a telemetry data error.

        Default implementation of Recorder.telemetry_data_issue() exporting the telemetry data error to all exporters of the recorder.
        See the docstrings of the Recorder interface for more details.
        """
        assert_that(not self._closed, "Recorder is already closed.")

        for exporter in self._exporters:
            self._export_telemetry_data_error(exporter, error_message, attributes)

    def _export_telemetry_data_error(self, exporter: Exporter, error_message: str, attributes: Optional[Attributes] = None) -> None:
        exporter_state = self._exporter_state[exporter]
        if exporter_state.telemetry_data_error_reported:
            return
        exporter_state.telemetry_data_error_reported = True
        error = TelemetryDataError.create(
            error_type=self._determine_telemetry_data_error_type(exporter_state), error_message=error_message, attributes=attributes
        )
        self._execute_export(exporter, exporter.export_telemetry_data_error, error)

    def _determine_telemetry_data_error_type(self, exporter_state: _ExporterState) -> TelemetryDataError.ErrorType:
        if exporter_state.any_data_exported_successfully:
            return TelemetryDataError.ErrorType.INCOMPLETE_TELEMETRY_DATA
        return TelemetryDataError.ErrorType.NO_TELEMETRY_DATA

    @safely_execute
    @override
    def close(self) -> None:
        """Close the recorder.

        Default implementation of Recorder.close() closing all exporters of the recorder.
        See the docstrings of the Recorder interface for more details.
        """
        # Whatever is left in the _spans dict, is an active span, we need to stop it.
        # Create a copy of the spans to avoid OrderedDict mutation during iteration
        active_spans = list(self._spans.values())
        for span in reversed(active_spans):
            self.stop(span)
        self._spans.clear()

        for exporter in self._exporters:
            try:
                exporter.close()
            except Exception as e:
                _logger.error(f"Could not close the exporter {exporter.__class__.__name__}: {e}")

        self._closed = True

    def get_active_spans_by_name(self, span_name: SpanName) -> List[Span]:
        """Get all activespans with the given name.

        Args:
            span_name: The name of the spans to find.

        Returns:
            List[Span]: A list of all spans with the given name. Note that in the most general case, nested spans can have the same name
            (but they will have different IDs and different parents). For example, if you define a span to cover the body of function foo()
            and foo() calls itself (directly or indirectly), while this function call is in progress,you will have two active spans with the same name.
        """
        return [span for span in self._spans.values() if span.name == span_name]

    def _should_export_span(self, span: Span) -> bool:
        """Check if the span should be exported based on the export customization mode and the span name filter."""
        if self._export_customization_mode == ExportCustomizationMode.EXPORT_ALL_SPANS:
            return True
        elif self._export_customization_mode == ExportCustomizationMode.WHITELIST_SPANS and span.name in self._span_name_filter:
            return True
        elif self._export_customization_mode == ExportCustomizationMode.BLACKLIST_SPANS and span.name not in self._span_name_filter:
            return True
        return False

    def _execute_export(self, exporter: Exporter, method: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
        """Execute the export method of the exporter.

        This method is responsible for executing export_xxx methods of the exporter and handles any errors that occur during the export
        in a specific way (see the class docstrings for more info).
        """
        if self._exporter_state[exporter].force_disabled:
            return
        exporter_state = self._exporter_state[exporter]
        try:
            method(*args, **kwargs)
            exporter_state.any_data_exported_successfully = True
        except Exception as e:
            exporter_state.num_errors += 1
            error_message = f"Error exporting data via exporter {exporter.__class__.__name__}: {e}"
            _logger.error(error_message)
            try:
                # call the exporter directly (not through _execute_export) as this is a special export call
                # to report an export issue. So we treat it differently (e.g., its success doesn't change
                # the any_data_exported_successfully flag)
                if not exporter_state.telemetry_data_error_reported:
                    exporter_state.telemetry_data_error_reported = True
                    exporter.export_telemetry_data_error(
                        TelemetryDataError.create(error_type=self._determine_telemetry_data_error_type(exporter_state), error_message=error_message)
                    )

            except Exception as e2:
                _logger.error(f"Could not export telemetry data error for exporter {exporter.__class__.__name__}: {e2}")
            # Note that we are not re-raising the original error. See class docstrings for more details.
        finally:
            if exporter_state.num_errors >= _DISABLE_EXPORTER_AFTER_N_FAILURES:
                error_message = f"Exporter {exporter.__class__.__name__} has been force disabled due to too many consecutive failures."
                _logger.error(error_message)
                exporter_state.force_disabled = True



# SPDX-License-Identifier: Apache-2.0
"""Telemetry configuration protocol for OneLogger."""

from typing import Dict, Optional, Protocol, Union, runtime_checkable

from strenum import StrEnum

from nv_one_logger.core.attributes import AttributeValue


class ApplicationType(StrEnum):
    """Enum for common application types."""

    # Model Training (can include validation and testing of model)
    TRAINING = "training"

    # Model Validation (without training)
    VALIDATION = "validation"

    # Batch Inference (inference on a batch of data)
    BATCH_INFERENCE = "batch_inference"

    # Online Inference (inference on a single data point)
    ONLINE_INFERENCE = "online_inference"

    # Data Processing (e.g., ETL, ELT, data ingestion or data transformation pipelines)
    DATA_PROCESSING = "data_processing"


@runtime_checkable
class TelemetryConfig(Protocol):
    """Protocol for telemetry configuration.

    This protocol defines the interface that any telemetry configuration must implement.
    It allows the core module to reference telemetry configs without creating circular dependencies
    and provides type safety while allowing different implementations for different application types.

    The @runtime_checkable decorator enables runtime type checking with isinstance() and issubclass().
    Without this decorator, isinstance(obj, TelemetryConfig) would always return False, even for
    objects that structurally implement all the required methods and properties.

    Example usage:
        # This works because of @runtime_checkable
        if isinstance(some_config, TelemetryConfig):
            print("Valid telemetry config")

    Note: @runtime_checkable has some limitations - it only checks for the presence of methods
    and properties, not their signatures or return types. It's a structural check, not a
    behavioral one.
    """

    # Flag indicating whether the application has training iterations
    @property
    def is_train_iterations_enabled(self) -> bool:
        """Whether the application has training iterations that OneLogger should track metrics for."""
        ...

    # Flag indicating whether the application has validation/evaluation iterations
    @property
    def is_validation_iterations_enabled(self) -> bool:
        """Whether the application has validation/evaluation iterations that OneLogger should track metrics for."""
        ...

    # Flag indicating whether the application has test iterations
    @property
    def is_test_iterations_enabled(self) -> bool:
        """Whether the application has test iterations that OneLogger should track metrics for."""
        ...

    # Flag indicating whether the application saves checkpoints
    @property
    def is_save_checkpoint_enabled(self) -> bool:
        """Whether the application saves checkpoints that OneLogger should track metrics for."""
        ...

    # Application type for this telemetry configuration
    @property
    def app_type(self) -> Union[ApplicationType, str]:
        """Get the application type for this telemetry configuration."""
        ...

    # Custom metadata specific to telemetry. This metadata will be logged
    # as attributes of telemetry-related spans and events.
    @property
    def custom_metadata(self) -> Optional[Dict[str, AttributeValue]]:
        """Custom metadata specific to telemetry configuration."""
        ...



# SPDX-License-Identifier: Apache-2.0
from abc import ABC, abstractmethod
from enum import Enum, auto

from overrides import EnforceOverrides, override

from nv_one_logger.core.event import ErrorEvent, Event, TelemetryDataError
from nv_one_logger.core.exceptions import OneLoggerError
from nv_one_logger.core.span import Span


class ExportError(OneLoggerError):
    """Error raised when exporting data fails."""

    def __init__(self, message: str):
        super().__init__(message)


class Exporter(ABC, EnforceOverrides):
    """Interface for exporting data from the instrumented application to an observability backend or a collector pipeline."""

    @abstractmethod
    def initialize(self) -> None:
        """Initialize the exporter.

        An exporter can be used only after being initialized.

        Raises:
            ExportError: If there is an error during the initialization operation.
        """

    @abstractmethod
    def export_start(self, span: Span) -> None:
        """Export a newly started span along with its current attributes and its start event.

        Args:
            span: The span to export. This span can be still in progress.

        Raises:
            ExportError: If there is an error during the export operation.
        """

    @abstractmethod
    def export_stop(self, span: Span) -> None:
        """Export a stopped/finished span along with its current attributes and its stop event.

        Args:
            span: The span to export. This span must be already stopped.

           This span must be already stopped.

        Raises:
            ExportError: If there is an error during the export operation.
        """

    @abstractmethod
    def export_event(self, event: Event, span: Span) -> None:
        """Export an event that occurred for an active span.

        Note: you do not need to call this method for start and stop events as they are exported in the export_start and export_stop methods.
        Args:
            event: The event to export.
            span: The span that the event belongs to. This span must be still active.

        Raises:
            ExportError: If there is an error during the export operation.
        """

    @abstractmethod
    def export_error(self, event: ErrorEvent, span: Span) -> None:
        """Export an error event that occurred for an active span.

        Args:
            event: The error event to export.
            span: The span that the error event belongs to. This span must be still active.

        Raises:
            ExportError: If there is an error during the export operation.
        """

    @abstractmethod
    def export_telemetry_data_error(self, error: TelemetryDataError) -> None:
        """Export a telemetry data error.

        This method is used to signal to the telemetry backend that the telemetry data is not complete/correct.

        Args:
            issue: The telemetry data issue to export.

        Raises:
            ExportError: If there is an error during the export operation.
        """

    @abstractmethod
    def close(self) -> None:
        """Shut down the exporter.

        Use this method to release any resources held by the exporter or flush any pending data that is not yet exported.
        Raises:
            ExportError: If there is an error during the export operation.
        """


class BaseExporter(Exporter):
    """Base class for all exporters that need to be initialized.

    This base class ensures that the exporter is initialized before it is used and not used after it is closed.
    """

    class _State(Enum):
        """State of the exporter."""

        NOT_INITIALIZED = auto()
        READY = auto()
        CLOSED = auto()

    def __init__(self) -> None:
        self._state: BaseExporter._State = self._State.NOT_INITIALIZED

    @override
    def initialize(self) -> None:
        """Initialize the exporter."""
        self._state = self._State.READY

    @override
    def export_start(self, span: Span) -> None:
        """Export a newly started span along with its current attributes and its start event."""
        self._check_state()

    @override
    def export_stop(self, span: Span) -> None:
        """Export a stopped/finished span along with its current attributes and its stop event."""
        self._check_state()

    @override
    def export_event(self, event: Event, span: Span) -> None:
        """Export an event that occurred for an active span."""
        self._check_state()

    @override
    def export_error(self, event: ErrorEvent, span: Span) -> None:
        """Export an error event that occurred for an active span."""
        self._check_state()

    @override
    def export_telemetry_data_error(self, error: TelemetryDataError) -> None:
        """Export a telemetry data error."""
        self._check_state()

    @override
    def close(self) -> None:
        """Shut down the exporter."""
        self._state = self._State.CLOSED

    @property
    def ready(self) -> bool:
        """Check if the exporter is in ready state.

        Returns:
            bool: True if the exporter is initialized and ready to use, False otherwise.
        """
        return self._state == self._State.READY

    @property
    def closed(self) -> bool:
        """Check if the exporter is in ready state.

        Returns:
            bool: True if the exporter is initialized and ready to use, False otherwise.
        """
        return self._state == self._State.CLOSED

    def _check_state(self) -> None:
        if self._state == self._State.NOT_INITIALIZED:
            raise ExportError("Exporter not initialized. Call initialize first.")
        if self._state == self._State.CLOSED:
            raise ExportError("Exporter is closed and cannot be used anymore.")



# SPDX-License-Identifier: Apache-2.0
from contextlib import contextmanager
from typing import Generator, Optional

from nv_one_logger.api.one_logger_provider import OneLoggerProvider
from nv_one_logger.api.recorder import Recorder
from nv_one_logger.core.attributes import Attributes
from nv_one_logger.core.exceptions import assert_that
from nv_one_logger.core.internal.logging import get_logger
from nv_one_logger.core.span import Span, SpanName

_logger = get_logger(__name__)


def get_recorder() -> Recorder:
    """Return the singleton recorder.

    This is a simply a shortcut for getting the recorder used by the timed_span context manager.
    """
    return OneLoggerProvider.instance().recorder  # type: ignore[no-any-return]


@contextmanager
def timed_span(
    name: SpanName,
    span_attributes: Optional[Attributes] = None,
    start_event_attributes: Optional[Attributes] = None,
) -> Generator[Span, None, None]:
    """
        Context manager for recording a span.

        Note that OneLoggerProvider.instance().configure() must be called once per process (e.g., at app start up) before using the timed_span context manager.
        See the "How to use One Logger" section of the README.md file for more information.

        Args:
            name: The name of the span.
            span_attributes: Optional attributes to add to the span.
            start_event_attributes: Optional attributes to add to the start event.

        Yields:
            the span


        Usage:
        On app start up:
        ```python
            from nv_one_logger.api.timed_span import get_recorder, timed_span

            def main():
                # One-time initialization of the library
                recorder = ... # You can use a factory that takes some config parameters and builds a recorder.
                            # Or in simple use cases, just use recorder = DefaultRecorder(exporters=[...])
                OneLoggerProvider.instance().configure(recorder)
        ```

        In other parts of the application:
        ```python
            from nv_one_logger.api.timed_span import get_recorder, timed_span

            def foo():
                with timed_span(name="foo", start_event_attributes=Attributes({...})):
                    # business logic for operation foo()
        ```

        or for more advanced use cases, you can access the recorder directly for recording events and errors:
        ```python

            def bar():
                with timed_span(name="application", span_attributes=Attributes({"app_tag": "some tag"})) as span:
                    # some business logic

                    # You can record events of interests within a timed span.
                    get_recorder().event(span, Event.create(...))

                    # Or record errors
                    if(...):
                        get_recorder().error(span, "Error in bar", ...)

                    # some more business logic

    ```

    """
    recorder = get_recorder()
    assert_that(recorder, "One Logger is not configured! Use OneLoggerProvider.instance().configure() to configure One Logger.")
    span = recorder.start(span_name=name, span_attributes=span_attributes, start_event_attributes=start_event_attributes)

    try:
        yield span
    except Exception as e:
        recorder.error(span, f"Error in {name}: ", e)
        raise e
    finally:
        recorder.stop(span)



# SPDX-License-Identifier: Apache-2.0
from enum import Enum


class RecordType(Enum):
    """The type of record to be used by backends that log to a file, e.g. FileBackend and LoggerBackend."""

    # The start of a span.
    START = "start"
    # The end of a span.
    STOP = "stop"
    # A complete span, that is the start and stop of a span are combined into a single record for efficiency.
    COMPLETE = "complete"
    # A single event that can occur anytime in a span, e.g. a training iteration in a training loop.
    EVENT = "event"
    # An error is a special event that is used to record an exception or error encountered in the application.
    APPLICATION_ERROR = "application_error"
    # An error is a special event that is used to record an exception or error encountered in one logger code.
    TELEMETRY_DATA_ERROR = "telemetry_data_error"



# SPDX-License-Identifier: Apache-2.0
from abc import ABC, abstractmethod
from typing import Optional

from overrides import EnforceOverrides

from nv_one_logger.core.attributes import Attributes
from nv_one_logger.core.event import ErrorEvent, Event
from nv_one_logger.core.span import Span, SpanName
from nv_one_logger.core.time import TracingTimestamp


class Recorder(ABC, EnforceOverrides):
    """An abstract base class for creating and recording of spans and events.

    The application can call the Recorder API to  start/stop spans, add events, or report errors.
    See the "How to use One Logger" section of the README.md file for more information.

    The Recorder is in charge of creating the Span/Event objects and then using one or more exporters
    to send the spans, events, and errors to to one or more backends.
    The recorder decides which exporters to use and when to call them.

    Here are a few examples of what can be done in a Recorder:
    * Filter (not export) certain events (based on attribute values or verbosity).
    # Handle errors occuring while exporting data to each backend.
    * Add new attributes to a span or events or even create new events. For example, if you have a long-running
    span for model training in which multiple "save checkpoint" events are emitted, and you want to keep track
    of the avg, max, and min save times across all the checkpoints, the receiver can keep track of all "save
    checkpoint" events so far and maintain avg, max, and min values of the "duration" attribute of those events
    and then emit a "check point stats update" event periodically.    See the README.md for more information.
    """

    @abstractmethod
    def start(
        self,
        span_name: SpanName,
        span_attributes: Optional[Attributes] = None,
        start_event_attributes: Optional[Attributes] = None,
        start_time: Optional[TracingTimestamp] = None,
        parent_span: Optional[Span] = None,
    ) -> Span:
        """
        Create a new span and record its start event.

        Args:
            name: The name of the span.
            span_attributes: Optional attributes to add to the span.
            start_event_attributes: Optional attributes to add to the start event.
            start_time: The start time of the span; the current timestamp will be used if not specified.
            parent_span: The parent span of the new span. If not specified, the new span will be created as a child of the latest active span
            (or will be a root span if there is no active span).
        Returns:
            The span created.
        """
        pass

    @abstractmethod
    def stop(
        self,
        span: Span,
        stop_event_attributes: Optional[Attributes] = None,
        stop_time: Optional[TracingTimestamp] = None,
    ) -> None:
        """
        Stop the span and record its stop event.

        Args:
            span: The span to stop.
            stop_event_attributes: Optional attributes to add to the stop event.
            stop_time: The stop time of the span; the current timestamp will be used if not specified.
        """
        pass

    @abstractmethod
    def event(self, span: Span, event: Event) -> Event:
        """
        Create a new event and record it immediately.

        The event is added to the last span that was started, unless a span is provided.

        Args:
            span: Span to add the event to.
            event: The event to record.
        Returns:
            The event recorded.
        """
        pass

    @abstractmethod
    def error(
        self,
        span: Span,
        error_message: str,
        exception: Optional[Exception] = None,
    ) -> ErrorEvent:
        """
        Create a new error event with the given error message and exception and record it immediately.

        Use this to report errors in the application. See ErrorEvent for more information.

        The error event is added to the last span that was started, unless a span is provided.

        Args:
            span: Span to add the error event to.
            error_message: The message of the error event.
            exception: The exception of the error event.
        Returns:
            The error event recorded.
        """
        pass

    @abstractmethod
    def telemetry_data_error(self, error_message: str, attributes: Optional[Attributes] = None) -> None:
        """Create a new telemetry data error and record it immediately.

        Use this to report issues that occurred while collecting telemetry data (issues in telemtry code or state rather than the application).
        See TelemetryDataError for more information.

        Args:
            error_message: A descriptive message about the telemetry data error.
            attributes: Optional attributes associated with the event. You can use this to add more information about the error.
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Close the recorder and release any resources.

        All spans that were previously started will be stopped and the stop events recorded at the current time.
        Once the recorder is closed, it cannot be used again.
        """
        pass



# SPDX-License-Identifier: Apache-2.0
"""Configuration for OneLogger."""

import logging
import os
from enum import Enum
from pathlib import Path
from typing import Callable, Dict, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, model_validator
from typing_extensions import Self

from nv_one_logger.api.telemetry_config import TelemetryConfig
from nv_one_logger.core.attributes import AttributeValue
from nv_one_logger.core.internal.utils import evaluate_value

# Note: We use plain Python logging here instead of OneLogger's get_logger() to avoid circular dependencies.
# The logging module (nv_one_logger.core.internal.logging) imports from this config module,
# so we cannot import get_logger here. This logger is used for configuration validation messages
# and will always log regardless of OneLogger's log_level settings, which is appropriate for
# important configuration warnings.
logger = logging.getLogger(__name__)

# Ensure the logger has a handler to display messages even if the application hasn't configured logging yet
# This is important because config validation happens early in the initialization process
# if not logger.handlers:
#    _handler = logging.StreamHandler()
#    _handler.setFormatter(logging.Formatter('%(levelname)s:%(name)s:%(message)s'))
#    logger.addHandler(_handler)
#    logger.setLevel(logging.INFO)


class OneLoggerErrorHandlingStrategy(Enum):
    """Enum for the error handling strategy for OneLogger.

    This enum determines what happens when OneLogger encounters a fatal error (e.g., an exception in the instrumentation code or
    a problem with the OneLogger state). This does NOT affect handling of errors occuring when communicating with the telemetry backends
    (i.e., exporter failures, which are handled by the Recorder). Rather, this is about handling user errors when configuring OneLogger
    or bugs in telemetry code (e.g., assertion/invariant violations or hitting an inconsistent state).

    Our recommendation is to use PROPAGATE_EXCEPTIONS for development scenarios where you want maximum visibility into errors,
    or DISABLE_QUIETLY_AND_REPORT_METRIC_ERROR for production scenarios where training should continue even if logging fails.

    Note: If you don't explicitly specify a strategy, OneLogger will automatically apply smart defaults:
    - Enabled ranks (enable_for_current_rank=True): PROPAGATE_EXCEPTIONS (for error visibility)
    - Disabled ranks (enable_for_current_rank=False): DISABLE_QUIETLY_AND_REPORT_METRIC_ERROR (for safety)

    Read the docstrings for DefaultRecorder for more details on how errors from exporters (e.g., communication errors with telemetry
    backends) are handled when using that recorder.
    """

    """Propagate the exceptions to the caller.

    Use this strategy if you are OK with instrumentation exceptions to crash the application (or are willing to catch and handle such exceptions).
    Note that all of exceptions from One Logger will be instances of `OneLoggerError` or a subclass of it. This allows you to identify those exceptions  # noqa: E501
    and react to them accordingly.
    The advantage of this strategy is that you get maximum visibility into the errors in the instrumentation code.
    """
    PROPAGATE_EXCEPTIONS = "propagate_exceptions"

    """Disable OneLogger silently and report metric errors to the telemetry backends.

    With this option, if instrumentation code encounters any errors, the library catches/suppresses the exception letting the
    application continue running, logs a single error to the application logs, and informs the telemtry backends that
    the telemetry data has errors.
    """
    DISABLE_QUIETLY_AND_REPORT_METRIC_ERROR = "disable_quietly_and_report_metric_error"


class LoggerConfig(BaseModel):
    """Configuration for how OneLogger logs its messages and errors."""

    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO"

    log_format: str = "%(name)s - %(levelname)s - %(message)s"

    # Path to the file where OneLogger INFO logs its messages.
    log_file_path_for_info: Union[Path, str] = "onelogger.log"

    # Path to the file where OneLogger ERROR logs its messages.
    log_file_path_for_err: Union[Path, str] = "onelogger.err"

    @model_validator(mode="after")
    def validate_config(self) -> Self:
        """Validate the logger config."""
        if self.log_level not in {"DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"}:
            raise ValueError(f"log_level must be one of {'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'}, got {self.log_level}")
        return self


class OneLoggerConfig(BaseModel):
    """Configuration for OneLogger."""

    # Pydantic model configuration to allow arbitrary types in field definitions.
    # This is necessary because the telemetry_config field uses a Protocol type (TelemetryConfig),
    # which is a custom type that Pydantic cannot automatically generate a schema for.
    # By setting arbitrary_types_allowed=True,
    # we tell Pydantic to accept any type for this field and skip schema validation for it.
    #
    model_config = ConfigDict(arbitrary_types_allowed=True)

    # The unique name for application. This name is used to identify the telemetry data related to various executions of
    # the same application in the OneLogger system (over time, across different machines/clusters, and across
    # different versions of the application).
    application_name: str

    # Number of processes participating in the telemetry collection.
    # This is a fundamental configuration that affects the entire application, not just training telemetry.
    world_size_or_fn: Union[int, Callable[[], int]]

    @property
    def world_size(self) -> int:
        """Number of processes participating in the telemetry collection."""
        return evaluate_value(self.world_size_or_fn)

    # session_tag (or callable to generate the tag). session_tag is used to used to identify jobs that together contribute
    # to the same task. This means the jobs are "logically" part of a single larger job (e.g.,
    # a long running job that is split into multiple jobs due to resource constraints or resuming after a failure).
    session_tag_or_fn: Union[str, Callable[[], str]]

    @property
    def session_tag(self) -> str:
        """Get the session tag.

        Returns:
            str: The evaluated session tag value.
        """
        return evaluate_value(self.session_tag_or_fn)

    # Flag (or callable to return flag) that indicates if this is a baseline run for comparison purposes.
    # A baseline run is a run that is used to set a performance baseline for future runs.
    is_baseline_run_or_fn: Union[bool, Callable[[], bool]] = False

    @property
    def is_baseline_run(self) -> bool:
        """Get the baseline run flag.

        Returns:
            bool: The evaluated baseline run flag value.
        """
        return evaluate_value(self.is_baseline_run_or_fn)

    # Custom metadata to be logged with the training telemetry data.
    # This metadata will be logged as-is, without any modification as an
    # attribute of the APPLICATION span.
    custom_metadata: Optional[Dict[str, AttributeValue]] = None

    # The strategy to use for handling errors in the instrumentation code.
    # If not explicitly provided, OneLogger applies defaults based on rank enablement:
    # - Enabled ranks (where OneLogger is active): Will propagate errors
    # - Disabled ranks (where OneLogger is inactive): Will suppress errors quietly
    # You can explicitly set a strategy to override this automatic behavior.
    # See the enum docstring for more details on each strategy and for our recommendations on how to set this value.
    error_handling_strategy: OneLoggerErrorHandlingStrategy = OneLoggerErrorHandlingStrategy.PROPAGATE_EXCEPTIONS

    # Whether to enable logging for the current rank in distributed training.
    # This controls whether OneLogger is enabled for the current process/rank.
    # In distributed training scenarios, you can set this to False for ranks where you don't want logging.
    enable_for_current_rank: bool = True

    # Configuration for the logger used for logging messages and errors from the telemetry code.
    logger_config: LoggerConfig = LoggerConfig()

    # Version (or callable to return version) of the data schema used for summarizing metrics.
    # If the schema of the data you collect changes over time, you can use this value to
    # keep track of which schema version is used for which run.
    summary_data_schema_version: Union[str, str] = "1.0.0"

    # Telemetry-specific configuration.
    # This field contains telemetry-specific settings and parameters.
    # It is optional and can be None if no telemetry is needed.
    #
    # Note: This field uses the TelemetryConfig Protocol type, which is why we need
    # arbitrary_types_allowed=True in the model_config above. The actual value must be
    # a concrete class that implements the TelemetryConfig protocol interface.
    telemetry_config: Optional[TelemetryConfig] = None

    @model_validator(mode="before")
    @classmethod
    def apply_error_handling_defaults(cls, data):
        """Apply defaults for error handling strategy based on rank enablement.

        This applies production-safe defaults by automatically using quiet error handling
        for disabled ranks when the user hasn't explicitly specified an error handling strategy.
        """
        if isinstance(data, dict):
            # Check if user explicitly provided error_handling_strategy
            user_set_error_handling = "error_handling_strategy" in data

            # Apply automatic quiet error handling for disabled ranks when not explicitly set by user
            if not user_set_error_handling and not data.get("enable_for_current_rank", True):  # Default to True if not specified
                rank = int(os.environ.get("RANK", 0))
                logger.warning(
                    f"OneLogger: Setting error_handling_strategy to DISABLE_QUIETLY_AND_REPORT_METRIC_ERROR "
                    f"for rank (rank={rank}) with OneLogger disabled. "
                    "To override: explicitly set error_handling_strategy parameter."
                )

                # Set quiet error handling for production safety
                data["error_handling_strategy"] = OneLoggerErrorHandlingStrategy.DISABLE_QUIETLY_AND_REPORT_METRIC_ERROR

                # Set log level to CRITICAL to suppress further OneLogger logs for disabled ranks
                if "logger_config" not in data:
                    data["logger_config"] = {}
                if isinstance(data["logger_config"], dict) and "log_level" not in data["logger_config"]:
                    data["logger_config"]["log_level"] = "CRITICAL"

        return data

    @model_validator(mode="after")
    def validate_config(self) -> Self:
        """Validate the OneLogger configuration.

        This validator ensures that:
        - application_name is not empty
        - world_size is set to a positive value
        - custom_metadata keys are valid strings (if provided)
        - telemetry_config implements the TelemetryConfig protocol (if provided)

        Returns:
            OneLoggerConfig: The validated configuration.

        Raises:
            ValueError: If the configuration is invalid.
        """
        if not self.application_name or not self.application_name.strip():
            raise ValueError("application_name cannot be empty or whitespace-only")

        if self.world_size <= 0:
            raise ValueError("world_size must be set to a positive value")

        if self.custom_metadata is not None:
            for key in self.custom_metadata.keys():
                if not key.strip():
                    raise ValueError("custom_metadata keys must be non-empty strings")

        return self



# SPDX-License-Identifier: Apache-2.0
"""
This package contains the API for the OneLogger.

The classes in this package are used to set up OneLogger and capture spans, events, and errors.
See README.md for more information.
"""

from nv_one_logger.api.config import LoggerConfig, OneLoggerConfig, OneLoggerErrorHandlingStrategy  # noqa: F401



# SPDX-License-Identifier: Apache-2.0
import traceback
from enum import Enum
from typing import Any, Dict, Optional, Union

from strenum import StrEnum
from typing_extensions import TypeAlias

from nv_one_logger.core.attributes import Attributes
from nv_one_logger.core.exceptions import assert_that
from nv_one_logger.core.time import TracingTimestamp


class StandardEventName(StrEnum):
    """List of standard event names. These events are commonly used for spans regardless of the domain/application type."""

    # An event representing the start of a span.
    SPAN_START = "span_start"
    # An event representing the stop of a span.
    SPAN_STOP = "span_stop"
    # An event representing an error occurring during the execution of the span.
    ERROR = "error"
    # An event representing a telemetry data error.
    TELEMETRY_DATA_ERROR = "telemetry_data_error"


class StandardEventAttributeName(StrEnum):
    """List of standard event attributes. These attributes are commonly used for events regardless of the domain/application type."""

    TIMESTAMP_MSEC = "timestamp_msec"
    ERROR_MESSAGE = "error_message"
    EXCEPTION_TYPE = "exception_type"
    EXCEPTION_MESSAGE = "exception_message"
    EXCEPTION_TRACEBACK = "exception_traceback"
    # The type of a TelemetryDataError.
    TELEMETRY_DATA_ERROR_TYPE = "telemetry_data_error_type"


# The name of an event can be a string or an Enum. Using an Enum is preferred as it avoids issues such
# as typos or inconsistencies in the event name. So instead of passing free form strings, simply define an enum
# that inherits from str and Enum.
EventName: TypeAlias = Union[Enum, str]


class Event:
    """
    A class representing a span event.

    A Span Event represents a meaningful, singular point in time during the Span's duration.
    Each event has a name, timestamp, and a set of attributes.
    See https://opentelemetry.io/docs/concepts/signals/traces/#span-events


    Attributes:
        name: The name of the event.
        timestamp: indicates when the event occurred.
        attributes: Optional dictionary of event attributes.
    """

    def __init__(
        self,
        name: EventName,
        timestamp: TracingTimestamp,
        attributes: Optional[Attributes] = None,
    ):
        """
        Initialize a new Event instance.

        Args:
            name (EventName): The name of the event. See comments on EventName for more details.
            timestamp (TracingTimestamp): The timestamp when the event occurred.
            attributes (Optional[Attributes]): Optional attributes associated with the event.
        """
        assert_that(timestamp, f"Event {name} must have a valid timestamp")
        self._name: EventName = name

        # If you are adding a new field here and you want it to be exported as part of exporting the event,
        # add it to the attributes dictionary too. This helps us avoid adding special code for exporting different fields
        # of the event class.
        # In other words, the fields here are meant to be used in the library code and the attributes are meant to be
        # exported.
        self._timestamp: TracingTimestamp = timestamp

        # We do not allow adding attributes to an event after it is created. This ensures that the event is immutable and
        # reduces the complexity of the export process (e.g., we won't need to deal with the case that an event has changed after being exported).
        self._attributes: Attributes = attributes if attributes is not None else Attributes()
        self._attributes.add(StandardEventAttributeName.TIMESTAMP_MSEC, self._timestamp.milliseconds_since_epoch)

    @property
    def name(self) -> EventName:
        """Get the name of the event."""
        return self._name

    @property
    def timestamp(self) -> TracingTimestamp:
        """Get the timestamp of the event."""
        return self._timestamp

    @property
    def attributes(self) -> Attributes:
        """Get the attributes associated with the event."""
        return self._attributes

    @property
    def name_str(self) -> str:
        """Get the name of the event as a string."""
        if isinstance(self.name, Enum):
            return self.name.value
        return self.name

    @classmethod
    def create(cls, name: EventName, attributes: Optional[Attributes] = None, timestamp: Optional[TracingTimestamp] = None) -> "Event":
        """
        Create a new event with the given name and attributes timestamped with the current time.

        Args:
            name (EventName): The name of the event. See comments on EventName for more details.
            attributes (Optional[Attributes]): Optional attributes associated with the event.
        """
        if timestamp is None:
            timestamp = TracingTimestamp.now()
        return cls(name=name, timestamp=timestamp, attributes=attributes)

    def to_json(self) -> Dict[str, Any]:
        """Convert the event to a JSON-compatible dictionary that can be passed to json.dumps.

        Returns:
            dict: a JSON-compatible dictionary representation of the event.
        """
        ret: Dict[str, Any] = {"name": self.name_str, "timestamp": self.timestamp.to_json()}
        if self.attributes:
            ret["attributes"] = self.attributes.to_json()
        return ret

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> "Event":
        """Create an Event instance from a JSON-compatible dictionary (e.g., the return value of json.loads).

        Args:
            data: Dictionary containing the Event data.

        Returns:
            Event: An Event created from the data.
        """
        assert_that(data, "Event data must be a dictionary")
        assert_that(data.get("name") and data.get("timestamp"), f"Event must have a name and a timestamp: {data}")
        name = data["name"]
        if name == StandardEventName.ERROR:
            return ErrorEvent.from_json(data)
        else:
            return cls(
                name=name,
                timestamp=TracingTimestamp.from_json(data["timestamp"]),
                attributes=Attributes.from_json(data["attributes"]) if "attributes" in data else Attributes(),
            )

    def __eq__(self, other: object) -> bool:
        """Compare this event with another event for equality.

        Args:
            other: The other event to compare with

        Returns:
            bool: True if the events are equal, False otherwise
        """
        if not isinstance(other, Event):
            return False
        return self.name == other.name and self.timestamp == other.timestamp and self.attributes == other.attributes

    def __hash__(self) -> int:
        """Compute a hash value for this event.

        Returns:
            int: A hash value based on the event's name, timestamp, and attributes
        """
        return hash((self.name, self.timestamp, self.attributes))


class ErrorEvent(Event):
    """An event that represents an error in the application and contains an error message or an exception.

    This can be used both for errors that are represented as a string (e.g. an error message)
    as well as errors that are represented as an exception.

    Error events include a few standard attributes:
    - StandardEventAttributes.ERROR_MESSAGE (str): A descriptive message about the error.
    - StandardEventAttributes.EXCEPTION_TYPE (str): The type of the exception if the error was associated with an exception. Not set otherwise.
    - StandardEventAttributes.EXCEPTION_MESSAGE (str): The message from the exception if the error was associated with an exception. Not set otherwise.
    - StandardEventAttributes.EXCEPTION_TRACEBACK (str): The traceback of the exception if the error was associated with an exception and a traceback
        is available. Not set otherwise.
    """

    def __init__(
        self,
        timestamp: TracingTimestamp,
        error_message: str,
        exception_type: Optional[str] = None,
        exception_message: Optional[str] = None,
        exception_traceback: Optional[str] = None,
        attributes: Optional[Attributes] = None,
    ):
        """
        Initialize a new ErrorEvent instance.

        Use the static factory method create() to create an instance instead of this constructor.

        Args:
            timestamp (TracingTimestamp): The timestamp when the error occurred.
            error_message (str): A descriptive message about the error. If the error is based on an exception,
            this can be the same as the exception message or be a more detailed description of the error.
            exception_type (Optional[str]): The type of the exception if applicable
            exception_message (Optional[str]): The message from the exception if applicable
            exception_traceback (Optional[str]): The traceback of the exception if applicable. If
            this is not provided when an exception is provided, the traceback will be captured using traceback.format_exc().
            attributes (Optional[Attributes]): Optional attributes associated with the error event
        """
        assert_that(not exception_type or exception_message, "exception_message must be provided when exception_type is set.")

        # Add a few standard attributes to the event.
        all_attributes = Attributes()
        all_attributes.add(StandardEventAttributeName.ERROR_MESSAGE, error_message)
        if exception_type:
            all_attributes.add(StandardEventAttributeName.EXCEPTION_TYPE, exception_type)
        if exception_message:
            all_attributes.add(StandardEventAttributeName.EXCEPTION_MESSAGE, exception_message)
        if exception_traceback:
            all_attributes.add(StandardEventAttributeName.EXCEPTION_TRACEBACK, exception_traceback)
        if attributes:
            all_attributes.add_attributes(attributes)

        super().__init__(StandardEventName.ERROR, timestamp, all_attributes)
        self.error_message = error_message
        self.exception_type = exception_type
        self.exception_message = exception_message
        self.exception_traceback = exception_traceback

        if exception_type is not None and self.exception_traceback is None:
            self.exception_traceback = traceback.format_exc()

    @classmethod
    def create(  # type: ignore[override]
        cls, error_message: str, exception: Optional[Exception] = None, exception_traceback: Optional[str] = None
    ) -> "ErrorEvent":
        """
        Create a new error event with the given error message, exception and traceback.

        Args:
            error_message (str): A descriptive message about the error. If the error is based on an exception,
            this can be the same as the exception message or be a more detailed description of the error.
            exception (Optional[Exception]): The exception object if applicable
            exception_traceback (Optional[str]): The traceback of the exception if applicable

        Returns:
            ExceptionEvent: A new error event instance
        """
        return cls(
            timestamp=TracingTimestamp.now(),
            error_message=error_message,
            exception_type=exception.__class__.__name__ if exception else None,
            exception_message=str(exception) if exception else None,
            exception_traceback=exception_traceback,
        )

    def to_json(self) -> Dict[str, Any]:
        """Serialize the ErrorEvent to a JSON-compatible dictionary.

        Returns:
            dict: JSON-serializable representation of the ExceptionEvent
        """
        ret = super().to_json()
        ret["error_message"] = self.error_message

        if self.exception_type is not None:
            assert_that(self.exception_traceback is not None, "ErrorEvent must have a traceback if an exception is provided")
            assert_that(
                self.exception_message is not None,
                "ErrorEvent must have an exception message if an exception is provided",
            )
            ret["exception_type"] = self.exception_type
            ret["exception_message"] = self.exception_message
            ret["exception_traceback"] = self.exception_traceback

        return ret

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> "ErrorEvent":
        """Convert the ErrorEvent to a JSON-compatible dictionary that can be passed to json.dumps.

        Returns:
            dict: a JSON-compatible dictionary representation of the ErrorEvent.
        """
        return cls(
            timestamp=TracingTimestamp.from_json(data["timestamp"]),
            error_message=data["error_message"],
            exception_type=data["exception_type"] if "exception_type" in data else None,
            exception_message=data["exception_message"] if "exception_message" in data else None,
            exception_traceback=data["exception_traceback"] if "exception_traceback" in data else None,
            attributes=Attributes.from_json(data["attributes"]) if "attributes" in data else Attributes(),
        )

    def __eq__(self, other: object) -> bool:
        """Compare this error event with another error event for equality.

        Args:
            other: The other error event to compare with

        Returns:
            bool: True if the error events are equal, False otherwise
        """
        if not isinstance(other, ErrorEvent):
            return False
        return (
            super().__eq__(other)
            and self.error_message == other.error_message
            and self.exception_type == other.exception_type
            and self.exception_message == other.exception_message
            and self.exception_traceback == other.exception_traceback
        )

    def __hash__(self) -> int:
        """Compute a hash value for this error event.

        Returns:
            int: A hash value based on the error event's attributes
        """
        return hash((super().__hash__(), self.error_message, self.exception_type, self.exception_message, self.exception_traceback))


class TelemetryDataError(Event):
    """An event that represents a telemetry data error.

    This can be used to report issues that occurred while collecting telemetry data. Note that the difference between
    TelemetryDataIssue and ErrorEvent is that the latter is for errors in the application itself. In other words,
    ErrorEvent refers to errors that would occur even if OneLogger is disabled whereas TelemetryDataIssue is used to
    represent errors in OneLogger itself or in data collected by OneLogger.

    TelemetryDataError events include a few standard attributes:
    - StandardEventAttributes.TELEMETRY_DATA_ERROR_TYPE (str): The type of the telemetry data error.
    - StandardEventAttributes.ERROR_MESSAGE (str): A descriptive message about the telemetry dataerror.
    """

    class ErrorType(StrEnum):
        """The type of a telemetry data error."""

        # Indicates that no telemetry data could be collected for this execution of the application. For example, when the telemetry code
        # fails and gets disabled before any data is collected. Note that we cannot guarantee that all such cases are detected and reported
        # because even reporting "NO_TELEMETRY_DATA_COLLECTED" may fail. So reporting this condition is done on a best effort basis.
        NO_TELEMETRY_DATA = "no_telemetry_data"

        # Indicates that the telemetry data is incomplete/unreliable/corrupted. For example, if telemetry code encounters
        # a fatal error when some data is already collected or exported to the telemetry backends. This type of error
        # signals the backend that any data collected for the current run of the application is not reliable and should not be
        # used in analysis.
        INCOMPLETE_TELEMETRY_DATA = "incomplete_telemetry_data"

    def __init__(
        self,
        timestamp: TracingTimestamp,
        error_type: ErrorType,
        error_message: str,
        attributes: Optional[Attributes] = None,
    ):
        """
        Initialize a new TelemetryDataIssue instance.

        Use the static factory method create() to create an instance instead of this constructor.

        Args:
            timestamp (TracingTimestamp): The timestamp when the telemetry data error occurred.
            error_type (TelemetryDataErrorType): The type of the telemetry data error.
            error_message (str): A descriptive message about the telemetry data error.
            attributes (Optional[Attributes]): Optional attributes associated with the telemetry data error.
        """
        # Add a few standard attributes to the event.
        all_attributes = Attributes()
        all_attributes.add(StandardEventAttributeName.ERROR_MESSAGE, error_message)
        all_attributes.add(StandardEventAttributeName.TELEMETRY_DATA_ERROR_TYPE, error_type.value)
        if attributes:
            all_attributes.add_attributes(attributes)

        super().__init__(StandardEventName.TELEMETRY_DATA_ERROR, timestamp, all_attributes)
        self.error_type = error_type
        self.error_message = error_message

    @classmethod
    def create(cls, error_type: ErrorType, error_message: str, attributes: Optional[Attributes] = None) -> "TelemetryDataError":  # type: ignore[override]
        """Create a new telemetry data error event.

        Args:
            error_type (TelemetryDataErrorType): The type of the telemetry data error.
            error_message (str): A descriptive message about the telemetry data error.
            attributes (Optional[Attributes]): Optional attributes associated with the telemetry data error.

        Returns:
            TelemetryDataError: A new telemetry data error event instance
        """
        return cls(timestamp=TracingTimestamp.now(), error_type=error_type, error_message=error_message, attributes=attributes)

    def to_json(self) -> Dict[str, Any]:
        """Serialize the TelemetryDataIssue to a JSON-compatible dictionary.

        Returns:
            dict: JSON-serializable representation of the TelemetryDataIssue
        """
        ret = super().to_json()
        ret["error_type"] = self.error_type.value
        ret["error_message"] = self.error_message
        return ret

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> "TelemetryDataError":
        """Convert the TelemetryDataIssue to a JSON-compatible dictionary that can be passed to json.dumps.

        Returns:
            dict: a JSON-compatible dictionary representation of the TelemetryDataIssue.
        """
        return cls(
            timestamp=TracingTimestamp.from_json(data["timestamp"]),
            error_type=cls.ErrorType(data["error_type"]),
            error_message=data["error_message"],
            attributes=Attributes.from_json(data["attributes"]) if "attributes" in data else Attributes(),
        )

    def __eq__(self, other: object) -> bool:
        """Compare this TelemetryDataIssue event with another TelemetryDataIssue event for equality.

        Args:
            other: The other TelemetryDataError event to compare with

        Returns:
            bool: True if the TelemetryDataError events are equal, False otherwise
        """
        if not isinstance(other, TelemetryDataError):
            return False
        return super().__eq__(other) and self.error_type == other.error_type and self.error_message == other.error_message

    def __hash__(self) -> int:
        """Compute a hash value for this TelemetryDataIssue event.

        Returns:
            int: A hash value based on the TelemetryDataError event's attributes
        """
        return hash((super().__hash__(), self.error_type, self.error_message))



# SPDX-License-Identifier: Apache-2.0
import logging

from nv_one_logger.api.config import LoggerConfig
from nv_one_logger.api.one_logger_provider import OneLoggerProvider


def get_logger(name: str) -> logging.Logger:
    """Initialize a Python logger based on the user configuration.

    Args:
        name: Name of the logger

    Returns:
        logging.Logger: Configured Python logger instance
    """
    logger = logging.getLogger(name)
    if OneLoggerProvider.instance().one_logger_ready:
        logger_config: LoggerConfig = OneLoggerProvider.instance().config.logger_config
        formatter = logging.Formatter(logger_config.log_format)
        fh_info = logging.FileHandler(logger_config.log_file_path_for_info)
        fh_info.setLevel(logging.INFO)
        fh_info.setFormatter(formatter)
        logger.addHandler(fh_info)

        fh_err = logging.FileHandler(logger_config.log_file_path_for_err)
        fh_err.setLevel(logging.ERROR)
        fh_err.setFormatter(formatter)
        logger.addHandler(fh_err)

        logger.setLevel(logger_config.log_level)

    return logger



# SPDX-License-Identifier: Apache-2.0
import json
import os
from importlib.metadata import entry_points
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import yaml

from nv_one_logger.core.internal.logging import get_logger
from nv_one_logger.exporter.exporter import Exporter
from nv_one_logger.exporter.exporter_config import ExporterConfig

_logger = get_logger(__name__)


class ExporterConfigManager:
    """Manages exporter configuration with priority-based loading.

    This class handles the discovery, loading, and merging of exporter configurations
    from multiple sources with a defined priority order:
    1. Direct configuration (highest priority)
    2. File configuration (medium priority)
    3. Package configuration (lowest priority)

    The class automatically discovers exporter configurations from installed packages
    using Python entry points and provides methods to build final configurations
    and create exporter instances.
    """

    # Entry point name for discovering exporter configurations
    ENTRY_POINT_NAME = "nv_one_logger.exporter_configs"

    def __init__(self):
        """Initialize the ExporterConfigManager.

        Automatically loads exporter configurations from installed packages
        using the entry point system. Any packages that register themselves
        under the ENTRY_POINT_NAME will have their configurations loaded
        and available for use.

        Raises:
                None: Exceptions during entry point loading are logged as warnings
                      and do not prevent initialization.
        """
        self.entry_point_exporter_configs = self._get_exporter_configs_from_entry_points()

    def generate_export_config(
        self,
        direct_config: Optional[Union[List[ExporterConfig], List[Dict[str, Any]]]] = None,
        config_file_path: Optional[str] = None,
    ) -> List[ExporterConfig]:
        """Generate export configuration using priority order.

        This method merges configurations from multiple sources according to the
        defined priority system. Higher priority configurations override lower
        priority ones, while different exporter types are combined.

        Args:
                direct_config: Configuration passed directly to the method. Can be:
                        - List[ExporterConfig]: A list of ExporterConfig objects
                        - List[Dict[str, Any]]: A list of dictionaries, each representing an exporter config

                        If a list of dictionaries is provided, it will be automatically
                        converted to a list of ExporterConfig objects. This has the highest priority
                        and will override any conflicting configurations from other sources.
                config_file_path: Path to a configuration file (YAML or JSON). If
                        provided, this file will be loaded and merged with other
                        configurations. If None, the system will look for configuration
                        files in the current working directory or use the environment
                        variable ONE_LOGGER_EXPORTER_CONFIG_PATH.

        Returns:
                List[ExporterConfig]: The merged configuration containing all
                exporters from all sources, with higher priority configurations
                taking precedence over lower priority ones.

        Note:
                The priority order is:
                1. Direct configuration (highest priority)
                2. File configuration (medium priority)
                3. Package configuration (lowest priority)

                For the same exporter class, higher priority configurations completely
                replace lower priority ones. For different exporter classes, all
                configurations are combined.
        """
        # Start with package/team configs (lowest priority)
        base_config = self._get_package_configs()
        if base_config:
            _logger.info(f"Loaded {len(base_config)} exporter config(s) from packages")

        # Override with file config (medium priority)
        file_config = self._get_file_config(config_file_path)
        if file_config:
            _logger.info(f"Loaded {len(file_config)} exporter config(s) from file")
            # File config completely replaces base config, rather than merging
            base_config = file_config

        # Override with direct config (highest priority)
        if direct_config:
            # Convert to list of ExporterConfig if needed
            if isinstance(direct_config, list) and direct_config and isinstance(direct_config[0], dict):
                # Handle list of dictionaries
                direct_config = self._build_exporter_configs(direct_config)
            base_config = self._merge_configs(base_config, direct_config)

        _logger.info(f"Final configuration contains {len(base_config)} exporter(s)")
        return base_config

    def create_exporters_from_config(self, exporters_config: List[ExporterConfig], training_telemetry_config: Optional[Any] = None) -> List[Exporter]:
        """Create exporter instances from configuration.

        This method takes a list of ExporterConfig objects and creates actual exporter
        instances from the configuration. Only enabled exporters are created,
        and any failures during creation are logged as warnings.

        Args:
                exporters_config: The list of ExporterConfig objects containing exporter
                        definitions. Each exporter in the configuration must have a
                        valid class_name that can be dynamically imported.
                training_telemetry_config: Optional training telemetry configuration that
                        may be passed to exporters that require it.

        Returns:
                List[Exporter]: A list of created exporter instances. The list may
                        be empty if no exporters are enabled or if all exporters fail
                        to create. Each exporter in the list is a valid Exporter instance
                        that can be used for telemetry data export.

        Note:
                - Only exporters with enabled=True are created
                - Failed exporter creation is logged as warnings and skipped
                - The method uses dynamic import to load exporter classes
                - Special handling is provided for FileExporter to convert
                  string paths to Path objects
        """
        exporters = []

        # Create exporters from config only (no default nv_exporter)
        for exporter_config in exporters_config:
            if not exporter_config.enabled:
                continue

            exporter = self._create_exporter_from_config(exporter_config, training_telemetry_config)
            if exporter:
                exporters.append(exporter)

        return exporters

    # =============================================================================
    # PRIVATE IMPLEMENTATION METHODS
    # =============================================================================

    def _get_exporter_configs_from_entry_points(self) -> Dict[str, Any]:
        """Load exporter configurations from entry points (package/team configs)."""
        configs = {}
        try:
            # Python 3.8 compatibility: entry_points() doesn't support group parameter
            # Use dictionary-style access instead
            eps = entry_points()
            if hasattr(eps, "select"):
                # Python 3.10+ style
                entry_points_group = eps.select(group=self.ENTRY_POINT_NAME)
            elif hasattr(eps, "get"):
                # Python 3.8-3.9 style - returns dict-like object
                entry_points_group = eps.get(self.ENTRY_POINT_NAME, [])
            else:
                # Fallback: empty list if entry_points() returns unexpected format
                entry_points_group = []

            for entry_point in entry_points_group:
                try:
                    config_class = entry_point.load()
                    configs[entry_point.name] = config_class
                    _logger.info(f"Loaded exporter config: {entry_point.name}")
                except Exception as e:
                    _logger.warning(f"Failed to load exporter config {entry_point.name}: {e}")
        except Exception as e:
            _logger.warning(f"Failed to load exporter configurations: {e}")
        return configs

    def _get_package_configs(self) -> List[ExporterConfig]:
        """Get configurations from installed package configs."""
        package_configs = []

        for config_name, config_class in self.entry_point_exporter_configs.items():
            try:
                # Get default config from package
                default_config = config_class.get_default_config()
                exporter_config = ExporterConfig(
                    class_name=default_config["class_name"],
                    config=default_config.get("config", {}),
                    enabled=default_config.get("enabled", True),
                )
                package_configs.append(exporter_config)
            except Exception as e:
                _logger.warning(f"Failed to get default config for {config_name}: {e}")

        return package_configs

    def _get_file_config(self, config_file_path: Optional[str] = None) -> Optional[List[ExporterConfig]]:
        """Load configuration from file (config_file_path, cwd, or env)."""
        # Determine file path using priority order
        file_path = self._find_config_file(config_file_path)
        if not file_path:
            return None

        try:
            with open(file_path, "r") as f:
                if file_path.suffix.lower() in [".yaml", ".yml"]:
                    config_data = yaml.safe_load(f)
                elif file_path.suffix.lower() == ".json":
                    config_data = json.load(f)
                else:
                    _logger.warning(f"Unsupported config file format: {file_path.suffix}")
                    return None

            # Convert to list of ExporterConfig
            if isinstance(config_data, dict) and "exporters" in config_data:
                # Handle the {"exporters": [...]} format from files
                return self._build_exporter_configs(config_data["exporters"])
            else:
                _logger.warning("Invalid config file format: expected dict with 'exporters' key")
                return None

        except Exception as e:
            _logger.warning(f"Failed to load config file {file_path}: {e}")
            return None

    def _find_config_file(self, config_file_path: Optional[str] = None) -> Optional[Path]:
        """Find configuration file using priority order."""
        # 1. Direct path provided
        if config_file_path:
            path = Path(config_file_path)
            if path.exists():
                return path

        # 2. Environment variable (higher priority than current working directory)
        env_path = os.getenv("ONE_LOGGER_EXPORTER_CONFIG_PATH")
        if env_path:
            path = Path(env_path)
            if path.exists():
                return path

        # 3. Current working directory
        cwd_files = [
            "one_logger_exporters_config.yaml",
            "one_logger_exporters_config.yml",
            "one_logger_exporters_config.json",
        ]

        for filename in cwd_files:
            path = Path.cwd() / filename
            if path.exists():
                return path

        return None

    def _merge_configs(self, base: List[ExporterConfig], override: List[ExporterConfig]) -> List[ExporterConfig]:
        """Merge configurations with override taking precedence."""
        # Merge exporters
        merged_exporters = {}

        # Add base exporters
        for exporter in base:
            merged_exporters[exporter.class_name] = exporter

        # Override with override exporters
        for exporter in override:
            if exporter.class_name in merged_exporters:
                _logger.info(f"Overriding exporter config for '{exporter.class_name}'")
            else:
                _logger.info(f"Adding new exporter config for '{exporter.class_name}'")
            merged_exporters[exporter.class_name] = exporter

        # Merge configs for same exporter types
        for class_name, exporter in merged_exporters.items():
            if class_name in [e.class_name for e in override]:
                # Deep merge configs
                base_exporter = next((e for e in base if e.class_name == class_name), None)
                if base_exporter:
                    merged_config = {**base_exporter.config, **exporter.config}
                    if merged_config != base_exporter.config:
                        _logger.info(f"Merged config for '{class_name}': {base_exporter.config} + {exporter.config} = {merged_config}")
                    exporter.config = merged_config

        return list(merged_exporters.values())

    def _build_exporter_configs(self, config_list: List[Dict[str, Any]]) -> List[ExporterConfig]:
        """Convert a list of dictionaries to a list of ExporterConfig."""
        exporters = []
        for config_dict in config_list:
            exporter_config = ExporterConfig(
                class_name=config_dict["class_name"],
                config=config_dict.get("config", {}),
                enabled=config_dict.get("enabled", True),
            )
            exporters.append(exporter_config)
        return exporters

    def _create_exporter_from_config(self, exporter_config: ExporterConfig, training_telemetry_config: Optional[Any] = None) -> Optional[Exporter]:
        """Create an exporter from configuration."""
        try:
            class_name = exporter_config.class_name
            config = exporter_config.config.copy()

            # Split the class name to get module and class
            if "." in class_name:
                module_name, class_name_part = class_name.rsplit(".", 1)
                module = __import__(module_name, fromlist=[class_name_part])
                exporter_class = getattr(module, class_name_part)
            else:
                # Assume it's a class name in the current module
                exporter_class = globals()[class_name]

            # Handle special cases for specific exporter types
            if class_name == "nv_one_logger.training_telemetry.v1_adapter.v1_compatible_wandb_exporter.V1CompatibleExporter":
                # V1CompatibleExporter needs one_logger_config and config dictionary
                if training_telemetry_config is None:
                    _logger.warning("V1CompatibleExporter requires one_logger_config but none was provided")
                    return None

                return exporter_class(one_logger_config=training_telemetry_config, config=config)
            elif "file_path" in config:
                # Convert string path to Path object for any exporter that uses file_path
                file_path_str = config.pop("file_path")
                config["file_path"] = Path(file_path_str)
                return exporter_class(**config)
            else:
                return exporter_class(**config)

        except (ImportError, AttributeError, KeyError) as e:
            _logger.warning(f"Failed to load exporter class {exporter_config.class_name}: {e}")
            return None
        except Exception as e:
            _logger.warning(f"Failed to create exporter {exporter_config.class_name}: {e}")
            return None



# SPDX-License-Identifier: Apache-2.0
from typing import Any


class OneLoggerError(Exception):
    """
    Represents an error occurring in the one logger library.

    This exception is raised when there is an error in the internal state (e.g., due to incorrect usage of the one logger library).
    Note that the config of the library has knobs to control the behavior of the library when an error occurs (propagate this error
    to the application or log and silently suppress it). See OneLoggerErrorHandlingStrategy.

    ************** NOTE **************
    If you are contributing to the One Logger source code, make sure all the exceptions you raise are instances of this class or its
    source code, make sure all the exceptions you raise are instances of this class or its
    subclasses. See the comments on OneLoggerErrorHandlingStrategy.PROPAGATE_EXCEPTIONS to understand why.
    """

    def __init__(self, message: str):
        super().__init__(message)


def assert_that(condition: Any, message: str) -> None:
    """Safely asserts that a condition is true.

    In general, observability code must not cause the application to crash. Because of this, instead of using an assert,
    we use this function to check invariants. If an invariant is violated, we log an error and throw a OneLoggerError.
    Since we have code to catch OneLoggerErrors and suppress them, this will not cause the application to crash.
    """
    if not condition:
        raise OneLoggerError(message)



# SPDX-License-Identifier: Apache-2.0
"""Contains the MultiWindowTimer class."""

from typing import Optional

from nv_one_logger.core.exceptions import OneLoggerError
from nv_one_logger.core.time import TracingTimestamp


class MultiWindowTimer:
    """
    A class for tracking timing windows and their statistics.

    This class allows for the tracking the duration of an operation that occurs over
    multiple timing windows. That is, the operation may stop and start (resume) multiple
    times with periods of inactivity (pause) between them. This timer keeps track
    of each time window in which the operation is active, and provides properties to
    access the total time, minimum, maximum, and average duration of all completed
    windows.

    Here is an example: Imagine that you have a long-running job that loads a batch of
    data from a data lake, transforms it, saves transformed data, then loads the next
    batch of data and repeats the same process (and this continues until all data is
    processed).

    If we are interested in knowing the total time spent on loading data as well as
    min/max/avg of loading each batch of data, we can use this class, call start()
    every time we start loading a batch of data, and call stop() every tine we finish
    loading a batch of data. In this example, a "window" is the duration of time that
    the job spends on loading a batch of data (the period between each subsequent
    call to start() and stop()).

    Note that this class is not thread-safe and also it assumes that your activity
    windows are not overlapping (you finish one window before starting another).
    """

    def __init__(self) -> None:
        """Initialize a new MultiWindowTimer."""
        # Time info of a few important windows.
        self._current_window_start: Optional[TracingTimestamp] = None
        self._first_window_start: Optional[TracingTimestamp] = None
        self._latest_window_end: Optional[TracingTimestamp] = None
        self._latest_window_duration_sec: float = 0.0

        # Number of windows so far (including both active and completed windows).
        self._total_window_count: int = 0

        # Total time across all windows (does not include the time between a stop call and the next start call as that time is not part of any window).
        self._total_time_sec: float = 0.0

        # Min/max/avg duration of a window.
        self._min_window_duration_sec: float = float("inf")
        self._max_window_duration_sec: float = 0.0
        self._avg_window_duration_sec: float = 0.0

    def start(self, start_time: Optional[TracingTimestamp] = None) -> None:
        """
        Start a new timing window.

        Args:
            start_time: The start time of the window. If None, the current time is used.

        Raises:
            OneLoggerError: If the timer is already active.
        """
        if not start_time:
            start_time = TracingTimestamp.now()

        if self._current_window_start:
            raise OneLoggerError("Cannot start timer since it is already active")

        self._current_window_start = start_time
        self._total_window_count += 1

        if self._first_window_start is None:
            self._first_window_start = self._current_window_start

    def stop(self, stop_time: Optional[TracingTimestamp] = None) -> None:
        """
        Stop the current timing window and update statistics.

        Args:
            stop_time: The end time of the window. If None, the current time is used.

        Raises:
            OneLoggerError: If the timer is not active.
        """
        if not stop_time:
            stop_time = TracingTimestamp.now()

        try:
            if not self._current_window_start:
                raise OneLoggerError("Cannot stop timer since it is not active")
            if stop_time.seconds_since_epoch < self._current_window_start.seconds_since_epoch:
                raise OneLoggerError(
                    "Cannot stop timer with a stop time that is before the start time."
                    + f" stop_time: {stop_time}, current_window_start: {self._current_window_start}"
                )

            self._latest_window_end = stop_time
            self._latest_window_duration_sec = stop_time.perf_counter_seconds - self._current_window_start.perf_counter_seconds

            self._min_window_duration_sec = (
                min(self._min_window_duration_sec, self._latest_window_duration_sec)
                if self._min_window_duration_sec != float("inf")
                else self._latest_window_duration_sec
            )
            self._max_window_duration_sec = max(self._max_window_duration_sec, self._latest_window_duration_sec)

            self._total_time_sec += self._latest_window_duration_sec
            self._avg_window_duration_sec = self._total_time_sec / self._total_window_count
        finally:
            self._current_window_start = None

    def reset(self) -> None:
        """Reset the timer."""
        self._current_window_start = None

        self._first_window_start = None
        self._latest_window_end = None
        self._latest_window_duration_sec = 0

        self._total_window_count = 0
        self._total_time_sec = 0.0
        self._min_window_duration_sec = float("inf")
        self._max_window_duration_sec = 0.0
        self._avg_window_duration_sec = 0.0

    @property
    def is_active(self) -> bool:
        """Get whether the timer is currently active.

        An active timer is one that has been started but not yet stopped.

        Returns:
            bool: True if the timer is active, False otherwise.
        """
        return self._current_window_start is not None

    @property
    def current_window_start(self) -> Optional[TracingTimestamp]:
        """Get the start time of the current timing window.

        Returns:
            Optional[TracingTimestamp]: The start time of the current window, or None if no window is active.
        """
        return self._current_window_start

    @property
    def total_window_count(self) -> int:
        """Get the total number of windows so far (including both active and completed windows)."""
        return self._total_window_count

    @property
    def total_time_sec(self) -> float:
        """Get the total time across all timing windows.

        Returns:
            float: The total time in seconds.
        """
        return self._total_time_sec

    @property
    def min_window_duration_sec(self) -> float:
        """Get the minimum duration of any timing window.

        Returns:
            float: The minimum window duration in seconds.
        """
        return self._min_window_duration_sec

    @property
    def max_window_duration_sec(self) -> float:
        """Get the maximum duration of any timing window.

        Returns:
            float: The maximum window duration in seconds.
        """
        return self._max_window_duration_sec

    @property
    def avg_window_duration_sec(self) -> float:
        """Get the average duration of all timing windows.

        Returns:
            float: The average window duration in seconds.
        """
        return self._avg_window_duration_sec

    @property
    def latest_window_duration_sec(self) -> float:
        """Get the duration of the most recently completed timing window.

        Returns:
            float: The duration of the latest window in seconds.
        """
        return self._latest_window_duration_sec

    @property
    def first_window_start(self) -> Optional[TracingTimestamp]:
        """Get the start time of the first timing window.

        Returns:
            Optional[TracingTimestamp]: The start time of the first window, or None if no window has been started.
        """
        return self._first_window_start

    @property
    def latest_window_end(self) -> Optional[TracingTimestamp]:
        """Get the end time of the latest timing window.

        Returns:
            Optional[TracingTimestamp]: The end time of the latest window, or None if no window has been started.
        """
        return self._latest_window_end

    def __eq__(self, other: object) -> bool:
        """Compare this timer with another timer for equality."""
        if not isinstance(other, MultiWindowTimer):
            return False
        return (
            self._current_window_start == other._current_window_start
            and self._first_window_start == other._first_window_start
            and self._latest_window_end == other._latest_window_end
            and self._latest_window_duration_sec == other._latest_window_duration_sec
            and self._total_window_count == other._total_window_count
            and self._total_time_sec == other._total_time_sec
            and self._min_window_duration_sec == other._min_window_duration_sec
            and self._max_window_duration_sec == other._max_window_duration_sec
            and self._avg_window_duration_sec == other._avg_window_duration_sec
        )

    def __hash__(self) -> int:
        """Compute a hash value for this timer."""
        return hash(
            (
                self._current_window_start,
                self._first_window_start,
                self._latest_window_end,
                self._latest_window_duration_sec,
                self._total_window_count,
                self._total_time_sec,
                self._min_window_duration_sec,
                self._max_window_duration_sec,
                self._avg_window_duration_sec,
            )
        )



# SPDX-License-Identifier: Apache-2.0
import uuid
from enum import Enum
from typing import Any, Dict, List, Optional, Union, cast

from strenum import StrEnum
from typing_extensions import TypeAlias

from nv_one_logger.core.attributes import Attribute, Attributes, AttributeValue
from nv_one_logger.core.event import Event, StandardEventName
from nv_one_logger.core.exceptions import assert_that
from nv_one_logger.core.time import Timer, TracingTimestamp


class StandardSpanName(StrEnum):
    """List of span names that are commonly used (across all domains) for spans."""

    # This span corresponds to the entire execution of the application (in other words, the main() function).
    APPLICATION = "application"


class StandardSpanAttributeName(StrEnum):
    """
    List of attributes that are commonly used for spans.

    Only include attributes that are generic (not specific to any domain) and are applicable to all spans.
    For domain specific attributes, you can define your own class that inherits from both str and Enum (preferred) or
    simply pass the attribute name as a string.
    """

    # The duration of a stopped span in milliseconds. This attribute is automatically added when the span is stopped.
    DURATION_MSEC = "duration_msec"

    #  The color assigned to the NVTX marker for this span. Use NVTXColor for the values.
    NVTX_COLOR = "nvtx_color"


class NVTXColor(StrEnum):
    """The color assigned to an NVTX marker."""

    UNSET = "unset"
    GREEN = "green"
    BLUE = "blue"
    YELLOW = "yellow"
    PURPLE = "purple"
    RAPIDS = "rapids"
    CYAN = "cyan"
    RED = "red"
    WHITE = "white"
    DARK_GREEN = "darkgreen"
    ORANGE = "orange"


# The name of a span can be a string or an Enum. Using an Enum is preferred as it avoids issues such
# as typos or inconsistencies in the event name. So instead of passing free form strings, simply define an enum
# that inherits from str and Enum.
SpanName: TypeAlias = Union[Enum, str]


class Span:
    """A class representing a span in a distributed tracing system.

    A span represents a unit of work or operation. It can contain events, and can have attributes associated with it.
    Spans are used to track the execution of operations and their relationships in a distributed system.
    See https://opentelemetry.io/docs/concepts/signals/traces/#spans.

    A span can have a set of events as well as attributes associated with it.
    In our library, we ensure that a start event is created when the span is created and a stop event when the span is stopped.
    """

    def __init__(
        self,
        id: uuid.UUID,
        name: SpanName,
        start_event: Event,
        parent_span: Optional["Span"] = None,
        span_attributes: Optional[Attributes] = None,
    ):
        """Initialize a new Span instance.

        Note: Prefer using the static create() function instead of directly creating an instance of this class.

        Args:
            id: A unique identifier for the span.
            name: The name of the span.
            start_event: The event that marks the beginning of the span.
            span_attributes: Optional attributes to add to the span. If you are passing a subclass of Attributes, we ensure that the subclass typeis preserved
            even if you add more attributes later.
        """
        assert_that(start_event, f"Span {name} does not have a start event.")
        assert_that(
            start_event and start_event.timestamp and start_event.timestamp.perf_counter_seconds and start_event.timestamp.perf_counter_seconds,
            f"Span {name} must have a start event with a valid timestamp but got {start_event.timestamp}",
        )

        self._id: uuid.UUID = id
        self._name: SpanName = name
        self._parent_span: Optional["Span"] = parent_span
        self._events: List[Event] = [start_event]

        # We keep track of which attributes were set upon the creation of the span vs the ones that were added or changed after the span was created.
        # This is useful in cases where we want to export the span and its initial attributes upon creation and then upon completion of the span,
        # we want to only export any new or changed attributes.
        if span_attributes is None:  # Don't change this to "not span_attributes". We don't want to change the type (subclass type) even if the dict is empty.
            span_attributes = Attributes()
        self._initial_attributes: Attributes = span_attributes
        # Let's make sure the type of the _updated_attributes field matches that of _initial_attributes (e.g., the same subclass of Attributes).
        self._updated_attributes: Attributes = type(self._initial_attributes)({})

        self._timer = Timer()
        self._timer.start(start_time=start_event.timestamp)

    @property
    def id(self) -> uuid.UUID:
        """Get the unique identifier of the span."""
        return self._id

    @property
    def name(self) -> SpanName:
        """Get the name of the span."""
        return self._name

    @property
    def parent_span(self) -> Optional["Span"]:
        """Get the parent span of the span or None if the span is a root span."""
        return self._parent_span

    @property
    def events(self) -> List[Event]:
        """Get the list of events associated with the span."""
        return self._events

    @property
    def attributes(self) -> Attributes:
        """Get the attributes associated with the span.

        This includes the attributes set during creation as well as any updated attributes.
        """
        # Make sure we preserve the subclass of the attributes.
        return type(self._initial_attributes).merge(self._initial_attributes, self._updated_attributes)

    @property
    def updated_attributes(self) -> Attributes:
        """Get the attributes associated with the span that were added or changed  after the span was created."""
        return self._updated_attributes

    @property
    def name_str(self) -> str:
        """Get the name of the span as a string."""
        return str(self.name)

    @property
    def start_event(self) -> Event:
        """Get the start event of the span."""
        assert_that(len(self.events) and self.events[0].name == StandardEventName.SPAN_START, "Span must have a start event.")
        return self.events[0]

    @property
    def stop_event(self) -> Optional[Event]:
        """Returns the stop event if the span has been stopped, otherwise returns None."""
        if not self._timer.running:
            assert_that(
                len(self.events) > 1 and self.events[-1].name == StandardEventName.SPAN_STOP,
                "Span must have a stop event if it has been stopped.",
            )
            return self.events[-1]
        return None

    @property
    def active(self) -> bool:
        """Check if the span is still active."""
        assert_that(
            self._timer.running or (len(self.events) > 1 and self.events[-1].name == StandardEventName.SPAN_STOP),
            f"inconsistent span state: time running: {self._timer.running}, events: {self.events}",
        )
        return self._timer.running

    @classmethod
    def create(
        cls,
        name: SpanName,
        parent_span: Optional["Span"] = None,
        span_attributes: Optional[Attributes] = None,
        start_event_attributes: Optional[Attributes] = None,
        start_time: Optional[TracingTimestamp] = None,
    ) -> "Span":
        """Create a new span with the given parameters.

        Args:
            name: The name of the span.
            parent_span: The parent span of the new span. If not specified, the new span will be a root span.
            start_time: The start time of the span; the current timestamp will be used if not specified.
            span_attributes: Optional attributes to add to the span.
            start_event_attributes: Optional attributes to add to the start event.
            start_time: The start time of the span; the current time will be used if not specified.

        Returns:
            A new Span instance initialized with the provided parameters.
        """
        if not start_time:
            start_time = TracingTimestamp.now()
        start_event = Event(name=StandardEventName.SPAN_START, attributes=start_event_attributes, timestamp=start_time)
        return cls(id=uuid.uuid4(), name=name, start_event=start_event, parent_span=parent_span, span_attributes=span_attributes)

    def stop(self, stop_event_attributes: Optional[Attributes] = None, stop_time: Optional[TracingTimestamp] = None) -> None:
        """Stop the span.

        Args:
            stop_event_attributes: Optional attributes to add to the stop event.
            stop_time: The stop time of the span; the current timestamp will be used if not specified.
        """
        if not self._timer.running:
            return  # already stopped
        if not stop_time:
            stop_time = TracingTimestamp.now()
        stop_event = Event(name=StandardEventName.SPAN_STOP, attributes=stop_event_attributes, timestamp=stop_time)
        self.add_event(stop_event)
        self._timer.stop(reset=True)
        self.add_attribute(StandardSpanAttributeName.DURATION_MSEC, _get_complete_span_duration_msec(self))

    def add_event(self, event: Event) -> Event:
        """Add an event to the span."""
        assert_that(self._timer.running, "Cannot add an event to a span that is not running.")
        self.events.append(event)
        return event

    def add_attribute(self, name: str, value: AttributeValue) -> Attribute:
        """
        Add an attribute to the span.

        Args:
            name: The name of the attribute.
            value: The value of the attribute.

        Returns:
            The attribute.
        """
        self._updated_attributes[name] = Attribute(name=name, value=value)
        return self._updated_attributes[name]

    def add_attributes(self, attributes: Attributes) -> None:
        """Add a dictionary of attributes to the span."""
        self._updated_attributes.update(attributes)

    def to_json(self) -> Dict[str, Any]:
        """Convert the span to a JSON-compatible dictionary that can be passed to json.dumps.

        Returns:
            dict: a JSON-compatible dictionary representation of the span.
        """
        ret: Dict[str, Any] = {"name": self.name_str, "id": self.id, "events": [event.to_json() for event in self.events]}
        if self.attributes:
            ret["attributes"] = self.attributes.to_json()
        return ret

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> "Span":
        """Create a Span instance from a JSON-compatible dictionary (e.g., the return value of json.loads).

        Args:
            data: Dictionary containing the Span data.

        Returns:
            Event: A Span created from the data.
        """
        assert_that(data, "Span data must be a dictionary")
        assert_that(data.get("name") and data.get("id"), f"Span must have a name and an id: {data}")
        assert_that(
            data.get("events") and isinstance(data.get("events"), list) and len(data.get("events")) > 0,
            f"Span must have a start event: {data}",
        )
        events = [Event.from_json(event) for event in data["events"]]
        span = cls(
            id=data["id"],
            name=data["name"],
            start_event=events[0],
            span_attributes=Attributes.from_json(data["attributes"]) if "attributes" in data else Attributes(),
        )
        for event in events[1:]:
            span.add_event(event)
        return span

    def duration_msec(self) -> int:
        """Get the duration of the span in milliseconds.

        This method can be called only after the span is stopped.
        """
        assert_that(not self.active and self.stop_event, f"Span must be running or stopped and have a stop event {self}.")
        assert_that(StandardSpanAttributeName.DURATION_MSEC in self.attributes, f"Span must have a duration attribute: {self}.")
        duration = self.attributes.get(StandardSpanAttributeName.DURATION_MSEC).value
        assert_that(isinstance(duration, int), f"Span must have a duration attribute: {self}.")
        return cast(int, duration)


def _get_complete_span_duration_msec(span: Span) -> int:
    """
    Get the duration of a complete span.

    Args:
        span: The span to get the duration of. This span must have a stop event.

    Returns:
        The duration of the span in milliseconds.
    """
    assert_that(span.stop_event, "Span must be stopped and have a stop event.")
    duration_sec = span.stop_event.timestamp.perf_counter_seconds - span.start_event.timestamp.perf_counter_seconds  # type: ignore[union-attr]
    return int(duration_sec * 1000)



# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass, field
from typing import Any, Dict


@dataclass
class ExporterConfig:
    """Configuration for a single exporter.

    This dataclass represents the configuration for a single exporter instance.
    It contains the information needed to create and configure an exporter.

    Attributes:
        class_name: The full module path to the exporter class (e.g.,
            "nv_one_logger.exporter.file_exporter.FileExporter"). This is used
            for dynamic import and instantiation of the exporter.
        config: A dictionary containing configuration parameters for the exporter.
            The contents depend on the specific exporter type. For example,
            FileExporter might have {"file_path": "/tmp/log.txt"}.
        enabled: Whether this exporter should be created and used. If False,
            the exporter will be skipped during creation.
    """

    class_name: str
    config: Dict[str, Any] = field(default_factory=dict)
    enabled: bool = True



# SPDX-License-Identifier: Apache-2.0
import time
from dataclasses import dataclass
from typing import Any, Dict, Optional

from nv_one_logger.core.exceptions import assert_that


@dataclass(frozen=True)
class TracingTimestamp:
    """
    A class uses to keep track of timestamp of an event.

    This class is useful for performance tracing, as it
    - is timezone agnostic.
    - is high resolution as it uses a performance counter for accurate calculation of the time elapsed.
    - has a regular timestamp for the wall clock time (useful for human readability).

    See https://superfastpython.com/time-time-vs-time-perf_counter/ for more details.
    """

    # The time in seconds since the epoch as a floating-point number. This is wall clock time and is same as time.time()
    _seconds_since_epoch: float

    # The value (in fractional seconds) of a performance counter, i.e. a clock with the highest available resolution to measure a short duration.
    # The reference point of the returned value is undefined, so that only the difference between the results of two calls is valid.
    _perf_counter_seconds: float

    @property
    def seconds_since_epoch(self) -> float:
        """
        Get the time in seconds since the epoch.

        Returns:
            float: The time since the epoch in fractional seconds.
        """
        return self._seconds_since_epoch

    @property
    def milliseconds_since_epoch(self) -> int:
        """
        Get the time in milliseconds since the epoch.

        Returns:
            float: The time since the epoch in milliseconds.
        """
        return int(self.seconds_since_epoch * 1000)

    @property
    def perf_counter_seconds(self) -> float:
        """
        Get the value of the performance counter in fractional seconds.

        Returns:
            float: The value of the performance counter in fractional seconds
        """
        return self._perf_counter_seconds

    @classmethod
    def now(cls) -> "TracingTimestamp":
        """
        Create a new TracingTimestamp instance representing the current time.

        Returns:
            TracingTimestamp: A new instance of TracingTimestamp with the current time
        """
        return cls(_seconds_since_epoch=time.time(), _perf_counter_seconds=time.perf_counter())

    @classmethod
    def for_timestamp(cls, timestamp_sec: float, perf_counter: Optional[float] = None, validate_timestamp: bool = True) -> "TracingTimestamp":
        """
        Create a new TracingTimestamp based on a given timestamp.

        Note that this function is menat for creating tracing timestamps, so in almost all cases, the timestamp should be very
        close to the current time. We use a threshold of 10 hours to allow for delayed reporting of spans/events for long running span.
        Enforcing this constraint helps us catch bugs in the code that use this function (e.g., using microseconds instead of fractional
        seconds or passing a perf counter value instead of a timestamp). If you have a legitimate case for creating a tracing timestamp
        that is not close to the current time, you can set validate_timestamp to false.

        Args:
            timestamp_sec (float): The timestamp in seconds since the epoch to create the TracingTimestamp from.

            perf_counter_seconds (float):Optional. The value of the performance counter in fractional seconds to create the TracingTimestamp from.
            If not provided, the perf counter value will be calculated based on the difference between the current wall clock time and the given timestamp.
            This may result in a slight loss of precision as it uses only the wall clock time to create the TracingTimestamp. If you need high precision,
            you need to know the value of time.time() as well as time.perf_counter() at the time of the event.

            validate_timestamp (bool): Whether to validate the timestamp. If set to true, the timestamp will be validated against the current time
            and rejected if it is not within 10 hours of the curent time. See the function's docstring for more details.
            Returns:
                TracingTimestamp: A new instance of TracingTimestamp created from the given timestamp.
        """
        if not perf_counter:
            # Let's guess the value of the perf counter at the time of the event. This should be fairly accurate
            # if the system clock has not been changed since the event. In general, the precision of the time.perf_counter() function
            # may be relatively higher than the precision of  clocks used by other time functions in the time module,
            # such as the time.time() function.
            perf_counter = time.perf_counter() - (time.time() - timestamp_sec)

        if validate_timestamp:
            threashold = 60 * 10  # 10 hours. This allows delayed reporting of spans/events for long running span.
            assert_that(timestamp_sec <= time.time() + threashold, "timestamp is in the future. This is meant to be the time in seconds since epoch.")
            assert_that(timestamp_sec >= time.time() - threashold, "timestamp is too old. This is meant to be the time in seconds since epoch.")
        return cls(_seconds_since_epoch=timestamp_sec, _perf_counter_seconds=perf_counter)

    def to_json(self) -> Dict[str, float]:
        """Convert the TracingTimestamp to a JSON-compatible dictionary that can be passed to json.dumps.

        Returns:
            dict: a JSON-compatible dictionary representation of the TracingTimestamp.
        """
        return {"seconds_since_epoch": self.seconds_since_epoch, "perf_counter_seconds": self.perf_counter_seconds}

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> "TracingTimestamp":
        """Create a TracingTimestamp instance from a JSON-compatible dictionary (e.g., the return value of json.loads).

        Args:
            data: Dictionary containing the TracingTimestamp data.

        Returns:
            Event: A TracingTimestamp created from the data.
        """
        assert_that(data, "TracingTimestamp data cannot be None.")
        assert_that(
            data.get("seconds_since_epoch") and data.get("perf_counter_seconds"),
            f"TracingTimestamp must have seconds_since_epoch and perf_counter_seconds: {data}",
        )

        return cls(_seconds_since_epoch=data["seconds_since_epoch"], _perf_counter_seconds=data["perf_counter_seconds"])

    def __hash__(self) -> int:
        """
        Compute the hash value of this TracingTimestamp.

        Returns:
            int: A hash value based on both the epoch seconds and performance counter seconds.
        """
        return hash((self._seconds_since_epoch, self._perf_counter_seconds))

    def __eq__(self, other: object) -> bool:
        """
        Compare this TracingTimestamp with another object for equality.

        Args:
            other: The object to compare with.

        Returns:
            bool: True if the other object is a TracingTimestamp with the same values,
                  False otherwise.
        """
        if not isinstance(other, TracingTimestamp):
            return False
        return self._seconds_since_epoch == other._seconds_since_epoch and self._perf_counter_seconds == other._perf_counter_seconds


class Timer:
    """
    A class representing the duration of a span/event.

    It can be used to measure the duration with high precision.
    """

    _start_time: Optional[TracingTimestamp] = None
    _elapsed_time_since_last_reset: float = 0.0
    _running: bool = False

    @property
    def running(self) -> bool:
        """Get the running state of the timer."""
        return self._running

    def start(self, start_time: Optional[TracingTimestamp] = None) -> None:
        """
        Start the duration measurement.

        Args:
            start_time: The time to start the duration measurement. If not provided, the current time will be used.
        """
        if not start_time:
            start_time = TracingTimestamp.now()
        assert_that(not self._running, "start called when the timer was already running")
        self._start_time = start_time
        self._running = True

    def stop(self, reset: bool = False) -> float:
        """
        Stop the duration measurement.

        Args:
            reset: Whether to reset the duration measurement. If set to true, a subsequent call to start()
            will reset the elapsed time to 0; otherwise, calling start() again will resume the timer and
            the elapsed time will be accumulated.

        Returns:
            The elapsed time since the start of the timer.
        """
        assert_that(self._running and self._start_time, "stop called when the timer was not running")
        self._running = False
        self._elapsed_time_since_last_reset += time.perf_counter() - self._start_time.perf_counter_seconds  # type: ignore[union-attr]
        if reset:
            ret = self._elapsed_time_since_last_reset
            self.reset()
            return ret
        else:
            return self._elapsed_time_since_last_reset

    def reset(self, start: bool = False) -> None:
        """
        Reset the duration and optionally re-start it.

        Args:
            start_time: The time to start the duration measurement, if start is True. If not provided, the current time will be used.
            start: Whether to start the duration measurement immediately.
        """
        self._start_time = None
        self._elapsed_time_since_last_reset = 0.0
        self._running = False
        if start:
            self.start()



# SPDX-License-Identifier: Apache-2.0
"""Contains utilities for safe execution of code.

This module provides utilities that help ensure that errors in onelogger library do not cause the instrumented application to crash.
"""

import logging
from functools import wraps
from typing import Any, Callable

from nv_one_logger.api.config import OneLoggerErrorHandlingStrategy
from nv_one_logger.api.one_logger_provider import OneLoggerProvider
from nv_one_logger.core.internal.logging import get_logger


def _get_logger():
    """Get logger lazily to ensure it picks up configuration.

    Falls back to plain Python logging if OneLoggerProvider is not configured.

    TODO: This is a short-term fix for timing issues where loggers are created before
    OneLoggerProvider.configure() is called. Long-term solution: implement a mechanism
    in logging.py to track and update all previously created loggers when configuration
    changes, so we don't need lazy initialization.
    """
    try:
        return get_logger(__name__)
    except (AttributeError, AssertionError, TypeError):
        # OneLoggerProvider not configured, mocked, or config invalid - fall back to plain Python logging
        return logging.getLogger(__name__)


# Depending on the user provided OneLoggerErrorHandlingStrategy, we may need to handle issues in collecting or exporting telemetry data  # noqa: E501
# in different ways. However, regardless of the strategy, we don't want to pollute the application logs with tens of errors related from telemetry code.  # noqa: E501
# This flag allows us to control how aggresively we log telemetry data issues to the application logs.
__telemetry_data_issue_already_logged = False


def exception_guard(func: Callable[..., Any]) -> Callable[..., Any]:  # noqa: C901
    """Handle exceptions from the wrapped function according to the configured OneLoggerErrorHandlingStrategy.

    This is a decorator function that wraps a given function to handle exceptions. All the telemetry code must be guarded with this decorator
    (directly or indirectly) so that the error handling strategy chosen by the user is respected.

    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        global __telemetry_data_issue_already_logged
        config = OneLoggerProvider.instance().config
        try:
            return func(*args, **kwargs)
        except Exception as e:
            error_message = f"Exception in telemetry code when calling {func.__name__}: {e}"
            try:
                if OneLoggerProvider.instance().one_logger_enabled:
                    # Inform all the exporters that the telemetry data has issues.
                    OneLoggerProvider.instance().recorder.telemetry_data_error(error_message)
            except Exception:
                pass  # not much we can do here.
            if config.error_handling_strategy == OneLoggerErrorHandlingStrategy.PROPAGATE_EXCEPTIONS:
                _get_logger().error(error_message)
                raise e
            elif config.error_handling_strategy == OneLoggerErrorHandlingStrategy.DISABLE_QUIETLY_AND_REPORT_METRIC_ERROR:
                if OneLoggerProvider.instance().one_logger_enabled:
                    # Extra check to ensure that we only log the issue in the application logs only once.
                    # This is not normally needed as with the DISABLE_QUIETLY_AND_REPORT_METRIC_ERROR strategy,
                    # we disable one logger after encourntering the first telemetry data error, which means there will
                    # be no further calls to one logger and no further telemetry data errors. But to be on the safe side
                    # we add this check to prevent filling the application logs with one logger errors even if
                    # disabling one logger doesn't take effect for some reason.
                    if not __telemetry_data_issue_already_logged:
                        __telemetry_data_issue_already_logged = True
                        _get_logger().error(
                            f"OneLogger encountered a fatal error. Disabling OneLogger and telemetry data collection. Error: {error_message}"
                        )  # noqa: E501

                    OneLoggerProvider.instance().force_disable_logging()
            else:
                # This should never happen. If it does, it means you have added a new error handling strategy without updating this function.
                _get_logger().error(f"Unrecognized error handling strategy {config.error_handling_strategy}")
                raise e

    return wrapper


def safely_execute(func: Callable[..., Any]) -> Callable[..., Any]:
    """Execute the given callable safely.

    In particular, this decorator ensures that:
    - the wrapped function is executed only if one_logger is enabled.
    - applies exception_guard: any errors encountered during the execution of the wrapped function
      is handled according to the configured error handling strategy.
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if OneLoggerProvider.instance().one_logger_enabled:
            return exception_guard(func)(*args, **kwargs)
        else:
            # Skipping execution because OneLogger is not enabled.
            _get_logger().warning(f"Skipping execution of {func.__name__} because OneLogger is not enabled.")
            return None

    return wrapper



# SPDX-License-Identifier: Apache-2.0
import json
import logging
import re
from typing import Any, Dict, List

from overrides import override  # type: ignore[ancereportUnknownVariableType]

from nv_one_logger.core.internal.logging import get_logger
from nv_one_logger.exporter.base_json_exporter import BaseJsonExporter
from nv_one_logger.exporter.exporter import ExportError

_logger = get_logger(__name__)


class LoggerExporter(BaseJsonExporter):
    """Exporter implementation that writes spans and using a python logger.

    Events are formatted as key=value pairs with hierarchical keys separated by dots.
    """

    def __init__(self, logger: logging.Logger):
        super().__init__()
        self._logger = logger

    @override
    def initialize(self) -> None:
        """Initialize the LoggerExporter.

        Extends the base class implementation for the LoggerExporter.
        See the docstrings of the Exporter interface for more details.
        """
        super().initialize()

    def _format_log(self, json_dict: Dict[str, Any], prefix: str = "") -> str:
        """
        Format the dictionary into a log string. The output string will be in the format of "[key1=value1 | key2=value2 | ...]".

        Nested dictionaries are formatted recursively, and will be in the format of "key1.key2=value where key1 is the top
        level key and key2 is the nested key in the nested dictionary".

        The prefix argument indicates if the dictionary is a top-level dictionary or a nested dictionary.
        Top level dictionaries are enclosed in square brackets, and have some additional properties like the process id and
        the record count added to the dictionary.

        Args:
            event_json (dict): The dictionary to format.
            prefix (str): The prefix indicates if the dictionary is a top-level dictionary or a nested dictionary.
                          This function is in fact called recursively to format nested dictionaries, and for nested dictionaries
                          the prefix is the key of the nested dictionary in the parent dictionary.

        Limitations:
            - dictionary keys cannot contain dots.
        Returns:
            str: The formatted log string.
        """
        items: List[str] = []
        for k, v in sorted(json_dict.items(), reverse=True):
            if "." in k:
                raise ExportError(f"keys cannot contain dots: {k}")

            if isinstance(v, dict):
                child_prefix = f"{prefix}.{k}" if prefix else k
                formatted_dict = self._format_log(json_dict=v, prefix=child_prefix)
                if formatted_dict:
                    items.append(formatted_dict)
            else:
                # We are not producing json but this ensures the strings are encoded with double quotes.
                encoded_value = json.dumps(v)
                if prefix:
                    items.append(f"{prefix}.{k}={encoded_value}")
                else:
                    items.append(f"{k}={encoded_value}")
        if prefix:
            return " | ".join(items) if items else ""
        else:
            return "[" + " | ".join(items) + "]" if items else ""

    @staticmethod
    def deserialize_log(text: str) -> Dict[str, Any]:
        """
        Unformat a log string back into a dictionary by reversing the format_log function.

        Refer to that functionfor details on the format of the log string.
        """
        # First extract all the characters that are in the top level square brackets.
        pattern = r"\[(.*)\]"
        matches = re.findall(pattern, text, re.DOTALL)
        if not matches:
            raise ExportError(f"Log string must be enclosed in square brackets: {text}")

        # Then split the text by the pipe character with a space on each side of the pipe.
        items = matches[-1].strip().rstrip("|").split(" | ")  # FIXME - exclude text inside double quotes

        ret: Dict[str, Any] = {}
        for item in items:
            k, v = item.split("=")  # FIXME - exclude text inside double quotes
            d = ret
            while "." in k:
                k1, k2 = k.split(".", 1)
                if k1 not in d:
                    d[k1] = {}
                k = k2
                d = d[k1]
            try:
                d[k] = json.loads(v)
            except Exception as e:
                _logger.error(f"Failed to deserialize log string: {e}")
                d[k] = v
        return ret

    @override
    def _write(self, json_dict: Dict[str, Any]) -> None:
        """Write the given JSON-compatible dictionary to the logger."""
        json_str = self._format_log(json_dict)
        self._logger.info(json_str)



# SPDX-License-Identifier: Apache-2.0
from threading import RLock
from typing import Any, ClassVar, Dict, Generic, Type, TypeVar, cast

T = TypeVar("T")  # Type variable for the class using SingletonMeta


class SingletonMeta(type, Generic[T]):
    """A metaclass that implements the singleton pattern.

    This metaclass ensures that only one instance of a class is created.
    The instance is stored in a class-level dictionary and returned on subsequent calls.

    Attributes:
        __lock: A lock to ensure thread safety when creating instances
        _instances: A dictionary mapping class types to their singleton instances
    """

    _lock: ClassVar[RLock] = RLock()  # Using RLock to prevent deadlock in dependent singletons
    _instances: ClassVar[Dict[Type[Any], Any]] = {}

    def __call__(cls: Type[T], *args: Any, **kwargs: Any) -> T:
        """Create or return the singleton instance of the class."""
        # Always prevent direct instantiation
        raise TypeError(f"{cls.__name__} cannot be instantiated directly. Use {cls.__name__}.instance() instead.")

    def instance(cls: Type[T]) -> T:
        """Return the singleton instance of the class.

        Args:
            cls: The class to create an instance of

        Returns:
            The singleton instance of the class
        """
        if cls not in SingletonMeta._instances:
            with SingletonMeta._lock:
                if cls not in SingletonMeta._instances:
                    # Create a new instance using type.__call__
                    instance = super().__call__()
                    SingletonMeta._instances[cls] = instance
        return cast(T, SingletonMeta._instances[cls])



# SPDX-License-Identifier: Apache-2.0
import json
import os
from io import TextIOWrapper
from pathlib import Path
from typing import Any, Dict, Optional

from overrides import override  # type: ignore[ancereportUnknownVariableType]

from nv_one_logger.core.internal.logging import get_logger
from nv_one_logger.exporter.base_json_exporter import BaseJsonExporter
from nv_one_logger.exporter.exporter import ExportError

_logger = get_logger(__name__)


class FileExporter(BaseJsonExporter):
    """Exporter implementation that writes spans and events to a file."""

    def __init__(self, file_path: Path):
        """Initialize the FileExporter with a file path.

        Args:
            file_path: The path to the file to write the spans and events to.
        """
        super().__init__()
        self._filepath: Path = file_path
        self._file: Optional[TextIOWrapper] = None

    @override
    def initialize(self) -> None:
        """Initialize the FileExporter.

        Extends the base class implementation for the FileExporter.
        """
        super().initialize()
        self._file = self._open_file()
        self._last_start_span = None

    @override
    def close(self) -> None:
        """Close the FileExporter.

        Extends the base class implementation for the FileExporter.
        """
        if not self._file:
            return
        super().close()
        # This should be done AFTER super.close() because super.close() may need to write the last record.
        with self._lock:
            if self._file:
                self._file.close()
                self._file = None

    def _open_file(self) -> TextIOWrapper:
        dir = os.path.dirname(self._filepath)
        if dir:
            os.makedirs(dir, exist_ok=True)
        return open(self._filepath, "w")

    @override
    def _write(self, json_dict: Dict[str, Any]) -> None:
        """Write the given JSON-compatible dictionary to the file."""
        if not self._file:
            raise ExportError("File exporter not initialized or already closed!")

        json_str = json.dumps(json_dict, separators=(",", ":"))
        with self._lock:
            try:
                self._file.write(json_str + "\n")
                self._file.flush()
            except Exception as e:
                _logger.error(f"Error writing record to file: {e}")



# SPDX-License-Identifier: Apache-2.0

"""A class that maintains some summary statistics for a numeric metric."""


from __future__ import annotations

from typing import Generic, Optional, TypeVar

T = TypeVar("T", int, float)


class MetricSummarizer(Generic[T]):
    """
    A class that maintains some summary statistics for a numeric metric.

    Type Parameters:
        T: The type of the metric value, must be either int or float
    """

    latest_value: Optional[T] = None
    min_value: Optional[T] = None
    max_value: Optional[T] = None
    avg_value: Optional[float] = None
    total_value: Optional[T] = None
    count: int = 0

    def add_value(self, value: T) -> None:
        """Add a value to the metric summarizer."""
        self.count += 1
        self.latest_value = value
        self.min_value = value if self.min_value is None else min(self.min_value, value)
        self.max_value = value if self.max_value is None else max(self.max_value, value)
        self.total_value = value if self.total_value is None else self.total_value + value
        self.avg_value = value if self.avg_value is None else float(self.total_value) / self.count

    def __hash__(self) -> int:
        """
        Compute the hash value of this MetricSummarizer instance.

        Returns:
            int: A hash value based on the current state of the summarizer
        """
        return hash((self.latest_value, self.min_value, self.max_value, self.avg_value, self.total_value, self.count))

    def __eq__(self, other: object) -> bool:
        """
        Compare this MetricSummarizer instance with another object for equality.

        Args:
            other: The object to compare with

        Returns:
            bool: True if the other object is a MetricSummarizer with the same state,
                 False otherwise
        """
        if not isinstance(other, MetricSummarizer):
            return False
        other_summarizer: MetricSummarizer[T] = other
        return (
            self.latest_value == other_summarizer.latest_value
            and self.min_value == other_summarizer.min_value
            and self.max_value == other_summarizer.max_value
            and self.avg_value == other_summarizer.avg_value
            and self.total_value == other_summarizer.total_value
            and self.count == other_summarizer.count
        )



# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union, cast

from typing_extensions import TypeAlias

from nv_one_logger.core.exceptions import OneLoggerError, assert_that

# It's best to keep this types as simple as possible to avoid difficulties with serialization
# at the point of sending the data to various backends. Backends such as wandb and OTEL only support primitive
# types and simple lists of primitives. So we need to ensure that the values we use here are
# easily serializable to primitive types.
# When adding new types, make sure to update the `to_json` and `from_json` methods and if needed,
# update various exporters to support the new type.
# Note: Event though the type hint here allows for any list of primitive values, we have validation
# code that expects all the elements of the list to be of the same type.
_PrimitiveValue: TypeAlias = Union[str, bool, int, float]
AttributeValue: TypeAlias = Union[_PrimitiveValue, List[_PrimitiveValue]]


@dataclass
class Attribute:
    """
    An attribute attached to a span or an event. Each attribute has a name and a single value.

    The timestamp of the attribute is considered to be the same as the timestamp of its parent (a span or an event).
    See https://opentelemetry.io/docs/specs/otel/common/#attribute for more information.
    """

    name: str
    value: AttributeValue

    def __post_init__(self) -> None:
        """Validate the attribute value."""
        if isinstance(self.value, list) and len(self.value) > 0:
            elelemt_type = type(self.value[0])
            if not all(isinstance(item, elelemt_type) for item in self.value):
                raise OneLoggerError(f"All elements of a list attribute must be of primitive type {type(self.value[0])} but got {type(self.value)}")


class Attributes(Dict[str, Attribute]):
    """A set of attributes for a single span or event."""

    def __init__(self, dictionary: Optional[Dict[str, AttributeValue]] = None) -> None:
        if dictionary:
            for name, value in dictionary.items():
                self.add(name, value)

    def add(self, name: str, value: AttributeValue) -> "Attributes":
        """Add a new metric to the set of attributes."""
        self.add_attribute(Attribute(name, value))
        return self

    def add_attribute(self, attribute: Attribute) -> Attribute:
        """Add a new metric to the set of attributes."""
        self[attribute.name] = attribute
        return self[attribute.name]

    def add_attributes(self, other_attributes: "Attributes") -> None:
        """Add a new metric to the set of attributes."""
        for attr in other_attributes.values():
            self.add_attribute(attr)

    def to_json(self) -> Dict[str, Any]:
        """Convert all attributes to a JSON-compatible dictionary that can be passed to json.dumps.

        Returns:
            dict: a JSON-compatible dictionary representation of all attributes.
        """
        return {attribute.name: attribute.value for attribute in self.values()}

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> "Attributes":
        """Create a Attributes instance from a JSON-compatible dictionary.

        Args:
            data: Dictionary containing the attributes data

        Returns:
            Attributes: New attributes collection created from the data
        """
        attributes = cls()
        if not data:
            return attributes
        for name, value in data.items():
            attributes.add(name, value)

        return attributes

    @classmethod
    def merge(cls, attributes: Optional["Attributes"] = None, other: Optional["Attributes"] = None) -> "Attributes":
        """Merge two sets of attributes (see the note about the return type below).

        Returns: The merged attributes will be returned as an instance of the class on which merge() was called.
        That is,
           Attributes.merge(attr1, attr2) returns an instance of Attributes

           Assuming TrainingLoopAttributes is a subclass of Attributes,
           TrainingLoopAttributes.merge(attr1, attr2) returns an instance of TrainingLoopAttributes. This is useful when
           you have a subclass of Attributes and you want to merge attributes of that subclass with some new attributes without
           upcasting to the base Attributes class.

           If you truly want to preserve the type of the attributes, you are better off using the add_attributes() method instead.
        """
        merged = cls()
        if attributes:
            for attribute in attributes.values():
                merged.add_attribute(attribute)

        if other:
            for attribute in other.values():
                merged.add_attribute(attribute)
        return merged

    def get_bool_value(self, attribute_name: str) -> Optional[bool]:
        """Get the value of an attribute as a boolean. Works both for required and optional attributes.

        Args:
            attribute_name: The name of the boolean attribute to get the value of.

        Returns:
            The value of the attribute as a boolean, or None if the attribute is not present.

        Raises:
            OneLoggerError: If the attribute value is not a boolean.
        """
        if attribute_name not in self.keys():
            return None
        val = self[attribute_name].value
        assert_that(isinstance(val, bool), f"Attribute {attribute_name} must be a boolean. Got {val}.")
        return cast(bool, val)

    def get_int_value(self, attribute_name: str) -> Optional[int]:
        """Get the value of an attribute as an integer. Works both for required and optional attributes.

        Args:
            attribute_name: The name of the integer attribute to get the value of.

        Returns:
            The value of the attribute as an integer, or None if the attribute is not present.

        Raises:
            OneLoggerError: If the attribute value is not an integer.
        """
        if attribute_name not in self.keys():
            return None
        val = self[attribute_name].value
        assert_that(isinstance(val, int), f"Attribute {attribute_name} must be an integer. Got {val}.")
        return cast(int, val)

    def get_float_value(self, attribute_name: str) -> Optional[float]:
        """Get the value of an attribute as a float. Works both for required and optional attributes.

        Args:
            attribute_name: The name of the float attribute to get the value of.

        Returns:
            The value of the attribute as a float, or None if the attribute is not present.

        Raises:
            OneLoggerError: If the attribute value is not a float.
        """
        if attribute_name not in self.keys():
            return None
        val = self[attribute_name].value
        assert_that(isinstance(val, float) or isinstance(val, int), f"Attribute {attribute_name} must be a float. Got {val}.")
        return cast(float, val)

    def get_str_value(self, attribute_name: str) -> Optional[str]:
        """Get the value of an attribute as a string. Works both for required and optional attributes.

        Args:
            attribute_name: The name of the string attribute to get the value of.

        Returns:
            The value of the attribute as a string, or None if the attribute is not present.

        Raises:
            OneLoggerError: If the attribute value is not a string.
        """
        if attribute_name not in self.keys():
            return None
        val = self[attribute_name].value
        assert_that(isinstance(val, str), f"Attribute {attribute_name} must be a string. Got {val}.")
        return cast(str, val)



# SPDX-License-Identifier: Apache-2.0
"""Provider for the OneLogger singleton instance.

This module provides a singleton class for managing the global OneLogger instance.
"""

from typing import Optional

from nv_one_logger.api.config import OneLoggerConfig
from nv_one_logger.api.recorder import Recorder
from nv_one_logger.core.exceptions import assert_that
from nv_one_logger.core.internal.singleton import SingletonMeta


class OneLoggerProvider(metaclass=SingletonMeta["OneLoggerProvider"]):
    """A singleton class provider for OneLogger.

    This singleton provides a global point of entry for the OneLogger library.

    Why do we need this singleton?
    - Implementations of Recorder are often stateful. So an application needs to create a single Recorder and use it everywhere.
    - The code that uses this library for instrumentation needs to get access to the config of the library and the recorder. Without
      a singleton, the application would need to be changed to plumb the instance of Recorder and the config everywhere. This makes
      the ligrary harder to adopt.


    This singleton needs to be configured once per process as follows on application start up:
    ```python
        config = OneLoggerConfig(....) # You can use a factory that takes a json or other representation of the configs and creates a OneLoggerConfig object.
        recorder = ... # You can use a factory that takes some config parameters and builds a recorder.
                   # Or in simple use cases, just use recorder = DefaultRecorder(exporters=[...])
        OneLoggerProvider.instance().configure(config, recorder)
    ```

    Once configured, the application can use the singleton:
    ```python
        # In other parts of the application
        with OneLoggerProvider.instance().recorder.start("my_span"):
            ... # code here will be considered part of the span and will be timed/recorded.

        or use timed_span context manager.
    ```
    """

    _recorder: Optional[Recorder] = None
    _config: Optional[OneLoggerConfig] = None

    """
    If set to True, the logging will be forced to be disabled effectively disabling onelogger library.
    This flag is useful when the user configures onelogger to enable logging but later on, we decide to
    disable logging (e.g., the library itself decides to disable logging due to encountering errors during export).
    """
    _logging_force_disabled: bool = False

    def configure(self, config: OneLoggerConfig, recorder: Recorder) -> None:
        """
        Set the recorder for the singleton.

        Args:
            recorder: The recorder to set.
        """
        assert_that(config, "config cannot be None.")
        assert_that(
            (not self._recorder or self._recorder is recorder) and (not self._config or self._config is config),
            "OneLoggerProvider already configured! You must call configure() once and only once.",
        )
        self._config = config
        self._recorder = recorder

    @property
    def recorder(self) -> Recorder:
        """Return the recorder."""
        assert_that(self._recorder, "You need to call OneLoggerProvider.instance().configure() once in your application before accessing the recorder.")
        return self._recorder  # type: ignore[return-value]

    @property
    def config(self) -> OneLoggerConfig:
        """Return the config."""
        assert_that(self._config, "You need to call OneLoggerProvider.instance().configure() once in your application before accessing the config.")
        return self._config  # type: ignore[return-value]

    @property
    def one_logger_ready(self) -> bool:
        """Check if the one_logger is ready to be used."""
        return self._recorder is not None and self._config is not None

    @property
    def one_logger_enabled(self) -> bool:
        """Check if the one_logger is ready to be used and logging is not forced to be disabled."""
        return self.one_logger_ready and not self._logging_force_disabled and self.config.enable_for_current_rank

    def force_disable_logging(self) -> None:
        """Force logging to be disabled effectively disabling onelogger library."""
        self._logging_force_disabled = True



# SPDX-License-Identifier: Apache-2.0
"""Contains utilities for the core module."""

import functools
import os
from contextlib import contextmanager
from typing import Any, Callable, Generator, Optional, TypeVar, Union


@contextmanager
def temporarily_modify_env(var_name: str, new_var: Optional[str] = None) -> Generator[None, None, None]:
    """Temporarily modify an environment variable. The original value is restored when the context manager is exited.

    Args:
        var_name: The name of the environment variable to modify.
        new_var: The new value to set the environment variable to.
    """
    original_value = os.environ.get(var_name)

    if new_var is not None:
        os.environ[var_name] = new_var
    else:
        os.environ.pop(var_name, None)

    try:
        yield
    finally:
        if original_value is not None:
            os.environ[var_name] = original_value
        elif new_var is not None:
            os.environ.pop(var_name, None)


_T = TypeVar("_T")


def evaluate_value(value: Union[_T, Callable[[], _T]]) -> _T:
    """Evaluate a value that could be either a direct value or a callable.

    Args:
        value: Either a direct value or a callable that returns a value.

    Returns:
        The evaluated value.
    """
    if callable(value):
        return value()
    return value


def patch_function(original_func: Callable[..., Any]) -> Callable[..., Any]:
    """Patches a given function by allowing a caller to provide a custom wrapper.

    Args:
        original_func: The original function to be patched.

    Returns:
        A new function with the patched behavior.


    Example usage:
    # original function
    def my_func(a, b):
        return a + b

    # Define a wrapper that adds logging and calls the original function
    def my_wrapper(original_func, *args, **kwargs):
        print(f"Calling {original_func.__name__} with arguments: {args}, {kwargs}")
        result = original_func(*args, **kwargs)  # Call the original function
        print(f"Result: {result}")
        return result

    # Patch the original function using the utility
    my_func = patch_function(my_func)(my_wrapper)

    # Calling my_func will now runs the extra logging logic in my_wrapper before calling the original function and prints the result after.
    result = my_func(3, 5)
    """

    def patch(wrapper_func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(original_func)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            # Call the wrapper provided by the caller
            return wrapper_func(original_func, *args, **kwargs)

        return wrapped

    return patch


def patch_method(original_method: Callable[..., Any]) -> Callable[..., Any]:
    """
    Patches a given object method by allowing a caller to provide a custom wrapper.

    Args:
        original_method: The original method to be patched.

    Returns:
        A new method with the patched behavior.

    # Usage Example:

    class MyClass:
        def __init__(self, name):
            self.name = name

        def greet(self, greeting):
            return f"{greeting}, {self.name}!"

    # Define a wrapper that adds logging and calls the original method
    def greet_wrapper(original_method: Callable, self: Any, *args: Any, **kwargs: Any) -> Any:
        print(f"Calling {original_method.__name__} on {self} with arguments: {args}, {kwargs}")
        result = original_method(self, *args, **kwargs)  # Call the original method with "self"
        print(f"Result: {result}")
        return result

    # Patch the greet method of MyClass using the utility
    MyClass.greet = patch_method(MyClass.greet)(greet_wrapper)

    # Test the patched method
    obj = MyClass("Alice")
    result = obj.greet("Hello")
    """

    def patch(wrapper_func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(original_method)
        def wrapped(self: Any, *args: Any, **kwargs: Any) -> Any:
            # Call the wrapper provided by the caller with "self" as the first argument
            return wrapper_func(original_method, self, *args, **kwargs)

        return wrapped

    return patch



# SPDX-License-Identifier: Apache-2.0
"""This module provides functions to get the version number of the package."""

from importlib.metadata import PackageNotFoundError, version
from pathlib import Path
from typing import Optional

import toml

from nv_one_logger.core.internal.logging import get_logger

_logger = get_logger(__name__)


def find_pyproject(start_dir: Optional[Path] = None) -> Optional[Path]:
    """Find the pyproject.toml file by traversing up the directory tree."""
    if start_dir is None:
        start_dir = Path(__file__).resolve().parent

    current_dir = Path(start_dir)
    while current_dir != current_dir.parent:
        pyproject_path = current_dir / "pyproject.toml"
        if pyproject_path.is_file():
            return pyproject_path
        current_dir = current_dir.parent

    return None


def get_version_from_pyproject() -> Optional[str]:
    """Get the version number from pyproject.toml file."""
    pyproject_path = find_pyproject()
    if pyproject_path is None:
        print("Error: pyproject.toml file not found.")
        return None

    try:
        with open(pyproject_path, "r") as f:
            data = toml.load(f)

        version = data.get("project", {}).get("version", None)
        if version:
            return str(version)

        version = data.get("tool", {}).get("poetry", {}).get("version", None)
        return str(version) if version else None

    except Exception as e:
        _logger.error(f"Error reading version from pyproject.toml: {str(e)}")
        return None


def get_version(pkg_name: str) -> str:
    """Get the version number from pyproject.toml file.

    NOTE: importlib.metadata only supports get version from
          installed package. so we just read the version number from toml file.
          see: https://github.com/python-poetry/poetry/issues/273#issuecomment-726059226

    :return: version number. {major}.{minor}.{patch}
    :rtype: str
    """
    # Try to get version from importlib.metadata
    try:
        return version(pkg_name)
    except PackageNotFoundError:
        pass

    # Try to get version from toml file
    ver = get_version_from_pyproject()
    if ver is None:
        _logger.error(f"Version not found for package {pkg_name}")
        return "unknown"
    return ver



# SPDX-License-Identifier: Apache-2.0
import os
from abc import abstractmethod
from threading import RLock
from typing import Any, Dict, Optional

from overrides import override  # type: ignore[ancereportUnknownVariableType]

from nv_one_logger.core.event import ErrorEvent, Event, TelemetryDataError
from nv_one_logger.core.exceptions import OneLoggerError, assert_that
from nv_one_logger.core.internal.logging import get_logger
from nv_one_logger.core.span import Span
from nv_one_logger.exporter.exporter import BaseExporter
from nv_one_logger.exporter.internal.record_type import RecordType

_logger = get_logger(__name__)


class BaseJsonExporter(BaseExporter):
    """A base class that exports spans and events and their attributes as JSON.

    Events are formatted as key=value pairs with hierarchical keys separated by dots.
    Each implementation can choose where to write/send the JSON data.
    """

    def __init__(self) -> None:
        """Initialize the BaseJsonExporter."""
        super().__init__()
        self._record_count: int = 0
        self._lock: RLock = RLock()
        self._pid: int = os.getpid()

        # This keeps track of the last span that was started but not yet written to the log.
        # This allows us to do an ioptimization: write a single record for a span that
        # has no children spans.
        # We only allow one span to be open but not written to the log at a time.
        self._last_start_span: Optional[Span] = None

    @override
    def initialize(self) -> None:
        """Initialize the BaseJsonExporter.

        Implementation of Exporter.initialize() for exporters that export the data as JSON.
        See the docstrings of the Exporter interface for more details.
        """
        super().initialize()
        self._last_start_span = None

    @override
    def export_start(self, span: Span) -> None:
        """Start the span and record its start event.

        Implementation of Exporter.export_start() for exporters that export the data as JSON.
        See the docstrings of the Exporter interface for more details.
        """
        super().export_start(span)

        if self._last_start_span is not None:
            self._record_span(self._last_start_span, RecordType.START)
        self._last_start_span = span

    @override
    def export_stop(self, span: Span) -> None:
        """Stop the span and record its stop event.

        Implementation of Exporter.export_stop() for exporters that export the data as JSON.
        See the docstrings of the Exporter interface for more details.
        """
        super().export_stop(span)

        if span.stop_event is None:
            raise OneLoggerError(f"Span {span.id} has no stop event")
        if self._last_start_span:
            if self._last_start_span.id == span.id:
                # Use a single record for a complete span
                # rather than two consecutive start and stop events.
                self._record_span(span, RecordType.COMPLETE)
                self._last_start_span = None
                return
            else:
                # record the previous span before recording the unrelated stop event.
                self._record_span(self._last_start_span, RecordType.START)
                self._last_start_span = None

        # record the stop event
        self._record_span(span, RecordType.STOP)

    @override
    def export_event(self, event: Event, span: Span) -> None:
        """Add an event to a span and record it.

        Implementation of Exporter.export_event() for exporters that export the data as JSON.
        See the docstrings of the Exporter interface for more details.
        """
        super().export_event(event, span)

        if self._last_start_span:
            self._record_span(self._last_start_span, RecordType.START)
            self._last_start_span = None
        json_dict = {
            "type": RecordType.EVENT.value,
            "span_id": str(span.id),
            "event": self._event_data_to_json(event),
        }
        self._export_json_dict(json_dict)

    @override
    def export_error(self, event: ErrorEvent, span: Span) -> None:
        """Add an error to a span and record it.

        Implementation of Exporter.export_error() for exporters that export the data as JSON.
        See the docstrings of the Exporter interface for more details.
        """
        super().export_error(event, span)

        if self._last_start_span:
            self._record_span(self._last_start_span, RecordType.START)
            self._last_start_span = None
        json_dict = {
            "type": RecordType.APPLICATION_ERROR.value,
            "span_id": str(span.id),
            "event": self._event_data_to_json(event),
        }
        self._export_json_dict(json_dict)

    @override
    def export_telemetry_data_error(self, error: TelemetryDataError) -> None:
        """Report a telemetry data error.

        Implementation of Exporter.export_telemetry_data_error() for exporters that export the data as JSON.
        See the docstrings of the Exporter interface for more details.
        """
        super().export_telemetry_data_error(error)

        json_dict = {
            "type": RecordType.TELEMETRY_DATA_ERROR.value,
            "error": self._event_data_to_json(error),
        }
        self._export_json_dict(json_dict)

    @override
    def close(self) -> None:
        """Close the exporter.

        Implementation of Exporter.close() for exporters that export the data as JSON.
        See the docstrings of the Exporter interface for more details.
        """
        if self._last_start_span:
            self._record_span(self._last_start_span, RecordType.START)
            _logger.warning(f"Span {self._last_start_span.id} was not closed")
            self._last_start_span = None
        super().close()

    def _record_span(self, span: Span, record_type: RecordType) -> None:
        """Record a span and its attributes as a JSON dictionary.

        Args:
            span: The span to record.
            record_type: The type of record to record.
        """
        json_dict: Dict[str, Any] = {
            "type": record_type.value,
            "id": str(span.id),
            "name": span.name_str,
        }
        if span.attributes:
            json_dict["attributes"] = span.attributes.to_json()

        if record_type == RecordType.START:
            json_dict["start_event"] = self._event_data_to_json(span.start_event)
        elif record_type == RecordType.COMPLETE:
            assert_that(span.stop_event, "cannot export a complete record type for an active span!")
            json_dict["start_event"] = self._event_data_to_json(span.start_event)
            json_dict["stop_event"] = self._event_data_to_json(span.stop_event)  # type: ignore[union-attr]
        elif record_type == RecordType.STOP:
            assert_that(span.stop_event, "missing stop event!")
            json_dict["stop_event"] = self._event_data_to_json(span.stop_event)  # type: ignore[union-attr]
        else:
            raise OneLoggerError(f"Invalid type: {record_type}")

        self._export_json_dict(json_dict)

    def _export_json_dict(self, json_dict: Dict[str, Any]) -> None:
        """Export a JSON-compatible dictionary after adding a few extra entries.

        Args:
            json_dict: The JSON-compatible dictionary to write.
        """
        with self._lock:
            try:
                self._increment_record_count()
                json_dict["count"] = self._record_count
                json_dict["pid"] = self._pid
                self._write(json_dict)
            except Exception as e:
                _logger.error(f"Error exporting json dict: {e}")
                raise e

    @abstractmethod
    def _write(self, json_dict: Dict[str, Any]) -> None:
        """Write a JSON-compatible dictionary to the output as a new record.

        Args:
            json_dict: The JSON-compatible dictionary to write.
        """
        raise NotImplementedError("Subclasses must implement this method!")

    def _increment_record_count(self) -> int:
        with self._lock:
            self._record_count += 1
            return self._record_count

    def _event_data_to_json(self, event: Event) -> Dict[str, Any]:
        """Convert the exportable data of an event to a JSON-compatible dictionary.

        Event.to_json is meant for general serialization of the event. This method, on the other hand,
        focuses on what we want to export about an event.
        """
        return {
            "name": event.name_str,
            "attributes": event.attributes.to_json(),
        }
